import os
import numpy as np
import math
import pandas as pd
import cv2
import pickle
import colorsys
import pyqtgraph as pg
import pyqtgraph.opengl as gl
import scipy.ndimage as ndi
from scipy.interpolate import interp1d, splprep, splev


def read_qss_file(qss_file_name):
    with open(qss_file_name, "r", encoding="UTF-8") as file:
        return file.read()


def check_loading_pickle_file(file_path):
    layer_dict, msg = None, None
    try:
        infile = open(file_path, "rb")
        layer_dict = pickle.load(infile)
        infile.close()
    except OSError:
        msg = "OSError: possible reason - disk full, please contact maintainers."
        return msg
    except (pickle.PickleError, pickle.UnpicklingError):
        msg = "Pickling error, please check your file or contact maintainers."
        return msg
    except EOFError:
        msg = "EOFError: possible reason - suspecting loading a broken file, please contact maintainers."
        return msg

    return layer_dict, msg


def read_excel_file(file_path):
    msg = None
    file_name, file_extension = os.path.splitext(file_path)
    if file_extension == ".csv":
        df = pd.read_csv(file_path)
    elif file_extension == ".xlsx":
        df = pd.read_excel(file_path)
    else:
        df = None
        msg = "Only CSV file and Excel file works."
    return df, msg


def read_label(file):
    lines = []
    for line in file:
        lines.append(line)

    n_lines = len(lines)
    label_index = np.zeros(n_lines - 14, "i")
    label_index[:] = np.nan
    label_colors = np.zeros((n_lines - 14, 3), "i")
    label_colors[:] = np.nan
    label_names = []
    for i in range(n_lines - 14):
        split_lines = lines[i + 14].split()
        label_index[i] = int(split_lines[0])
        label_colors[i] = np.array(
            [split_lines[1], split_lines[2], split_lines[3]]
        ).astype(int)
        split_lines2 = lines[i + 14].split('"')
        label_names.append(split_lines2[1])

    return label_index, label_names, label_colors


def rotation_x(theta):
    ct = np.cos(theta)
    st = np.sin(theta)
    rx = np.array([[1, 0, 0], [0, ct, -st], [0, st, ct]])
    return rx


def rotation_y(theta):
    ct = np.cos(theta)
    st = np.sin(theta)
    ry = np.array([[ct, 0, st], [0, 1, 0], [-st, 0, ct]])
    return ry


def rotation_z(theta):
    ct = np.cos(theta)
    st = np.sin(theta)
    rz = np.array([[ct, -st, 0], [st, ct, 0], [0, 0, 1]])
    return rz


def d2td3(pos2d, ax, ay, o):
    pos3d = pos2d[0] * ax + pos2d[1] * ay + o
    return pos3d


def get_region_label(data, label_data, bregma):
    region_label = []
    for i in range(len(data)):
        temp = data[i] + bregma
        temp = temp.astype(int)
        region_label.append(label_data[temp[0], temp[1], temp[2]])
    return region_label


def get_region_label_info(region_label, label_info):
    unique_label = np.sort(np.unique(region_label))
    label_names = []
    label_acronym = []
    label_color = []
    region_count = []
    for i in range(len(unique_label)):
        # print(unique_label[i])
        if unique_label[i] == 0:
            label_names.append(" ")
            label_acronym.append(" ")
            label_color.append((128, 128, 128))
        else:
            da_ind = np.where(label_info["index"] == unique_label[i])[0][0]
            label_names.append(label_info["label"][da_ind])
            label_acronym.append(label_info["abbrev"][da_ind])
            label_color.append(label_info["color"][da_ind])

        region_count.append(len(np.where(np.ravel(region_label) == unique_label[i])[0]))

    return region_count, label_names, label_acronym, label_color, unique_label


def calculate_virus_info(data_list, pieces_names, label_data, label_info, bregma):
    temp_data = data_list[0]
    for i in range(1, len(data_list)):
        temp_data = np.vstack([temp_data, data_list[i]])

    data = temp_data.astype(int)

    # data = np.array([vox_data[0]])
    # for i in range(1, len(vox_data)):
    #     if np.any(vox_data[i] != data[-1]):
    #         data = np.vstack([data, vox_data[i]])

    region_label = get_region_label(data, label_data, bregma)
    unique_region = np.sort(np.unique(region_label))
    region_volume = []
    for c_region in unique_region:
        region_volume.append(len(np.where(label_data == c_region)[0]))

    # print(region_volume)
    (
        region_count,
        label_names,
        label_acronym,
        label_color,
        unique_label,
    ) = get_region_label_info(region_label, label_info)
    # print(region_count)

    res_dict = {
        "object_name": "virus",
        "data": data_list,
        "pieces_names": pieces_names,
        "label_id": unique_label,
        "label_name": label_names,
        "label_acronym": label_acronym,
        "label_color": label_color,
        "region_volume": region_volume,
        "virus_volume": region_count,
    }

    return res_dict


def calculate_cells_info(data_list, pieces_names, label_data, label_info, bregma):
    data = data_list[0]
    for i in range(1, len(data_list)):
        data = np.vstack([data, data_list[i]])

    region_label = get_region_label(data, label_data, bregma)
    (
        region_count,
        label_names,
        label_acronym,
        label_color,
        unique_label,
    ) = get_region_label_info(region_label, label_info)

    res_dict = {
        "object_name": "cell",
        "pieces_names": pieces_names,
        "data": data_list,
        "label_name": label_names,
        "label_acronym": label_acronym,
        "label_color": label_color,
        "region_count": region_count,
    }
    return res_dict


def calculate_drawing_info(data_list, pieces_names, label_data, label_info, bregma):
    data = data_list[0]
    for i in range(1, len(data_list)):
        data = np.vstack([data, data_list[i]])
    # print(data)
    if "area" in pieces_names[0]:
        plot_mode = "area"
    else:
        plot_mode = "line"

    region_label = get_region_label(data, label_data, bregma)
    (
        region_count,
        label_names,
        label_acronym,
        label_color,
        unique_label,
    ) = get_region_label_info(region_label, label_info)

    res_dict = {
        "object_name": "drawing",
        "pieces_names": pieces_names,
        "data": data_list,
        "label_name": label_names,
        "label_acronym": label_acronym,
        "label_color": label_color,
        "region_count": region_count,
        "plot_mode": plot_mode,
    }
    return res_dict


def order_contour_pnt(pnt):
    order_ind = []
    x_min = np.min(pnt[:, 0])
    left_ind = np.where(pnt[:, 0] == x_min)[0]
    if len(left_ind) > 1:
        low_ind = np.where(pnt[left_ind, :] == np.min(pnt[left_ind, :]))[0]
        left_ind = left_ind[low_ind]
    left_pnt = pnt[left_ind, :]
    lower_inds = np.where(pnt[:, 1] <= left_pnt[1])[0]
    lower_pnts = pnt[:, lower_inds]


def calculate_contour_line(data):
    data = np.asarray(data)
    res = splprep([data[:, 0], data[:, 1], data[:, 2]], s=2)
    tck = res[0]
    # x_knots, y_knots, z_knots = splev(tck[0], tck)
    u_fine = np.linspace(0, 1, len(data))
    x_fine, y_fine, z_fine = splev(u_fine, tck)
    pnts = np.stack([x_fine, y_fine, z_fine], axis=1)
    # print(pnts)
    return pnts


def hex2rgb(hex):
    if "#" in hex:
        hex = hex.lstrip("#")
        rgb_color = [int(hex[i : i + 2], 16) for i in (0, 2, 4)]
    else:
        if len(hex) == 6:
            rgb_color = [int(hex[i : i + 2], 16) for i in (0, 2, 4)]
    return rgb_color[0], rgb_color[1], rgb_color[2]


# @jit()
def hsv2rgb(h, s, v):
    # h [0, 1], s, v [0, 1]
    h = h * 360
    c = v * s
    m = v - c
    x = c * (1 - np.abs(np.mod(h / 60, 2) - 1))
    if 0 <= h < 60:
        r, g, b = (c + m, x + m, m)
    elif 60 <= h < 120:
        r, g, b = (x + m, c + m, m)
    elif 120 <= h < 180:
        r, g, b = (m, c + m, x + m)
    elif 180 <= h < 240:
        r, g, b = (m, x + m, c + m)
    elif 240 <= h < 300:
        r, g, b = (x + m, m, c + m)
    elif 300 <= h < 360:
        r, g, b = (c + m, m, x + m)
    else:
        r, g, b = (m, m, m)
    r = r * 65535
    b = b * 65535
    g = g * 65535
    return r, g, b


def color_img(h, s, v):
    r, g, b = hsv2rgb(h, s, v)
    da_image = cv2.merge((r, g, b))
    return da_image


def merge_channels_into_single_img(czi_img, channel_colors):
    merged_img = np.zeros((czi_img.shape[0], czi_img.shape[1], 3))
    frac = 1 / len(channel_colors)
    # print(frac)
    for i in range(len(channel_colors)):
        temp_v = czi_img[:, :, i] / 65535
        temp_h = channel_colors[i][0]
        temp_s = channel_colors[i][1]
        # print(temp_h, temp_s)
        da_img = color_img(temp_h, temp_s, temp_v)
        merged_img = merged_img + frac * da_img
    return merged_img


def make_color_lut(channel_color: tuple, bit_level: int):
    r, g, b = colorsys.hsv_to_rgb(channel_color[0], channel_color[1], channel_color[2])
    colors = [(0, 0, 0), (r * 255, g * 255, b * 255)]
    color_map = pg.ColorMap(pos=[0, 1], color=colors)
    da_lut = color_map.getLookupTable(nPts=bit_level, mode=pg.ColorMap.FLOAT)
    da_lut = da_lut * 255
    return da_lut


def get_qhsv_from_czi_hsv(hsv_color: tuple):
    """
    Get QColor HSV color format (360, 255, 255) from czi HSV color format (1, 1, 1).
    :param hsv_color: (float [0., 1.], float [0., 1.], float [0., 1.])
    :return: QColor
    """
    h_val = int(hsv_color[0] * 360)
    s_val = int(hsv_color[1] * 255)
    v_val = int(hsv_color[2] * 255)
    da_color = (h_val, s_val, v_val)
    return da_color


def gamma_line(input, lims, gamma, depth_level):
    inv_gamma = 1.0 / gamma
    y = np.zeros(len(input))
    inds = np.logical_and(input >= lims[0], input <= lims[1])
    y[inds] = (
        np.power((input[inds] - lims[0]) / (lims[1] - lims[0]), gamma) * depth_level
    )
    y[input <= lims[0]] = 0
    y[input >= lims[1]] = depth_level
    return y


def crop_landscape(image, dim):
    r = (dim[0] / image.shape[0]) / (dim[0] / dim[1])
    nw = int(image.shape[1] * r)

    resized = cv2.resize(image, (nw, int(dim[1])), interpolation=cv2.INTER_AREA)

    half_width = int(dim[0]) / 2
    half_shape_width = int(resized.shape[1]) / 2

    start_x = half_shape_width - half_width
    end_x = half_width + half_shape_width
    cropped = resized[0 : dim[1], start_x:end_x]

    return cropped


def crop_portrait(image, dim):
    r = dim[1] / image.shape[1] / (dim[1] / dim[0])
    nh = int(image.shape[0] * r)

    resized = cv2.resize(image, (int(dim[0]), nh), interpolation=cv2.INTER_AREA)
    half_height = int(dim[1]) / 2
    half_shape_height = int(resized.shape[0]) / 2

    start_y = half_shape_height - half_height
    end_y = half_height + half_shape_height
    cropped = resized[start_y:end_y, 0 : dim[0]]

    return cropped


def create_other_size(image, file_name, dim, location):
    # 1 => width index, 0 => height index
    if image.shape[0] > image.shape[1]:
        cropped = crop_portrait(image, dim)
    else:
        cropped = crop_landscape(image, dim)

    cv2.imwrite(os.path.join(location, file_name), cropped)


def make_hist_data(image_data, max_val):
    hist_data_list = []
    for i in range(image_data.shape[2]):
        if np.max(image_data[:, :, i]) == 0:
            da_bins = max_val
        else:
            da_bins = np.max(image_data[:, :, i])
        hist_y, x = np.histogram(image_data[:, :, i], bins=da_bins)
        y = np.log1p(hist_y)
        y = y / np.max(y) * max_val
        y = np.append(y, 0)
        sfunc = interp1d(x, y, "cubic")
        inter_x = np.linspace(np.min(x), np.max(x), 200)
        inter_y = sfunc(inter_x)
        inter_y[inter_y < 0] = 0
        hist_data_list.append([inter_x, inter_y])
    return hist_data_list


def rect_contains(rect, point):
    if point[0] < rect[0]:
        return False
    elif point[1] < rect[1]:
        return False
    elif point[0] > rect[2] + rect[0]:
        return False
    elif point[1] > rect[3] + rect[1]:
        return False
    return True


def get_warp_matrix(src_tri_pnts, dst_tri_pnts):
    # Given a pair of triangles, find the affine transform.
    warp_mat = cv2.getAffineTransform(
        np.float32(src_tri_pnts), np.float32(dst_tri_pnts)
    )
    return warp_mat


def apply_affine_transform(src_img, warp_mat, size_dst):
    # Apply the Affine Transform just found to the src image
    dst = cv2.warpAffine(
        src_img.astype(np.float32),
        warp_mat,
        (size_dst[0], size_dst[1]),
        None,
        flags=cv2.INTER_LINEAR,
        borderMode=cv2.BORDER_REFLECT_101,
    )
    return dst


def warp_triangle(img1, img2, t1, t2, is_rgb=False):
    # Find bounding rectangle for each triangle
    r1 = cv2.boundingRect(t1.astype(np.float32))
    r2 = cv2.boundingRect(t2.astype(np.float32))

    # Offset points by left top corner of the respective rectangles
    t1_rect = t1 - r1[:2]
    t2_rect = t2 - r2[:2]
    t2_rect_int = t2_rect.astype(int)

    # Get mask by filling triangle
    if is_rgb:
        mask = np.zeros((r2[3], r2[2], img1.shape[2]), dtype=np.float32)
        cv2.fillConvexPoly(
            mask, np.int32(t2_rect_int), tuple(np.repeat(1.0, img1.shape[2])), 16, 0
        )
    else:
        mask = np.zeros((r2[3], r2[2]), dtype=np.float32)
        cv2.fillConvexPoly(mask, np.int32(t2_rect_int), 1, 16, 0)

    # Apply warpImage to small rectangular patches
    img1_rect = img1[r1[1] : r1[1] + r1[3], r1[0] : r1[0] + r1[2]]
    # img2Rect = np.zeros((r2[3], r2[2]), dtype = img1Rect.dtype)

    size = (r2[2], r2[3])

    warp_mat = get_warp_matrix(t1_rect, t2_rect)
    img2_rect = apply_affine_transform(img1_rect, warp_mat, size)
    img2_rect = img2_rect * mask

    # Copy triangular region of the rectangular patch to the output image
    yr = (r2[0], r2[0] + r2[2])
    xr = (r2[1], r2[1] + r2[3])
    if is_rgb:
        img2[xr[0] : xr[1], yr[0] : yr[1]] = img2[xr[0] : xr[1], yr[0] : yr[1]] * (
            tuple(np.repeat(1.0, img1.shape[2])) - mask
        )
    else:
        img2[xr[0] : xr[1], yr[0] : yr[1]] = img2[xr[0] : xr[1], yr[0] : yr[1]] * (
            1 - mask
        )
    img2[xr[0] : xr[1], yr[0] : yr[1]] = img2[xr[0] : xr[1], yr[0] : yr[1]] + img2_rect


def warp_points(pnts, t1, t2):
    # Find bounding rectangle for each triangle
    # r1 = cv2.boundingRect(t1.astype(np.float32))
    # r2 = cv2.boundingRect(t2.astype(np.float32))
    # print(r1)
    # print(r2)

    # pnts =

    da_pnts = np.hstack([pnts, np.ones((len(pnts), 1))])
    # Offset points by left top corner of the respective rectangles
    # t1_rect = t1 - r1[:2]
    # t2_rect = t2 - r2[:2]

    warp_mat = get_warp_matrix(t1, t2)
    output = np.dot(warp_mat, da_pnts.T).T

    return output


# calculate delanauy triangle
def calculateDelaunayTriangles(rect, points):
    # create subdiv
    subdiv = cv2.Subdiv2D(rect)

    # Insert points into subdiv
    for p in points:
        subdiv.insert(p)

    triangleList = subdiv.getTriangleList()

    delaunayTri = []

    pt = []

    for t in triangleList:
        pt.append((t[0], t[1]))
        pt.append((t[2], t[3]))
        pt.append((t[4], t[5]))

        pt1 = (t[0], t[1])
        pt2 = (t[2], t[3])
        pt3 = (t[4], t[5])

        if (
            rect_contains(rect, pt1)
            and rect_contains(rect, pt2)
            and rect_contains(rect, pt3)
        ):
            ind = []
            for j in range(0, 3):
                for k in range(0, len(points)):
                    if (
                        abs(pt[j][0] - points[k][0]) < 1.0
                        and abs(pt[j][1] - points[k][1]) < 1.0
                    ):
                        ind.append(k)
            if len(ind) == 3:
                delaunayTri.append((ind[0], ind[1], ind[2]))

        pt = []

    return delaunayTri


def get_vertex_ind_in_triangle(subdiv):
    triangles = subdiv.getTriangleList()
    n_triangles = len(triangles)
    tri_vet_inds = []
    for i in range(n_triangles):
        da_triangle = triangles[i]
        p1 = [da_triangle[0], da_triangle[1]]
        p2 = [da_triangle[2], da_triangle[3]]
        p3 = [da_triangle[4], da_triangle[5]]
        tri_vet_inds.append(
            [subdiv.locate(p1)[2], subdiv.locate(p2)[2], subdiv.locate(p3)[2]]
        )
    tri_vet_inds = np.asarray(tri_vet_inds) - 4
    return tri_vet_inds


def get_pnts_triangle_ind(tri_vet_inds, tri_data, size, pnts):
    # import cv2
    # import numpy as np
    # img_rec = (0, 0, 100, 200)
    # da_triangle = np.array([[0, 0], [0, 50], [100, 100]])
    # size = (200, 100)
    # pnts = np.array([[0, 1], [0, 2], [60, 200]])

    update_pnts = pnts.copy()
    n_pnts = len(pnts)
    loc = np.zeros(n_pnts)
    loc[:] = np.nan
    da_order = []

    ct_list = []
    for i in range(len(tri_vet_inds)):
        da_inds = tri_vet_inds[i]
        da_triangle = np.array(
            [tri_data[da_inds[0]], tri_data[da_inds[1]], tri_data[da_inds[2]]]
        )
        mask = np.zeros(size, dtype=np.uint8)
        cv2.fillPoly(mask, pts=[da_triangle], color=255)
        ct, hc = cv2.findContours(
            image=mask, mode=cv2.RETR_TREE, method=cv2.CHAIN_APPROX_NONE
        )
        ct_list.append(ct[0])

        # range_y = (np.min(da_triangle[:, 1]), np.max(da_triangle[:, 1]))
        # range_x = (np.min(da_triangle[:, 0]), np.max(da_triangle[:, 0]))
        #
        # valid_pnts_ind = [ind for ind in range(n_pnts) if range_x[0] <= pnts[ind][0] <= range_x[1] and range_y[0] <= pnts[ind][1] <= range_y[1]]
        # valid_pnts_ind = [ind for ind in valid_pnts_ind if ind not in da_order]
        # valid_pnts = pnts[valid_pnts_ind]
        # for j in range(len(valid_pnts)):
        #     res = cv2.pointPolygonTest(ct_list[i], (int(pnts[j][0]), int(pnts[j][1])), True)
        #     if res >= 0:
        #         loc.append(i)
        #         da_order.append(valid_pnts_ind[j])

    #     temp = np.zeros(len(update_pnts))
    #
    #
    # for i in range(len(tri_vet_inds)):
    #     da_
    #
    for i in range(len(pnts)):
        for j in range(len(ct_list)):
            da_ct = ct_list[j]
            res = cv2.pointPolygonTest(da_ct, (int(pnts[i][0]), int(pnts[i][1])), False)
            if res >= 0:
                loc[i] = j
                break

    return loc


def get_sides_points(img_size):
    size0 = img_size[1] - 1
    size1 = img_size[0] - 1
    side_lines = np.asarray(
        [
            [[0, 0], [size0, 0]],
            [[size0, 0], [size0, size1]],
            [[0, size1], [size0, size1]],
            [[0, 0], [0, size1]],
        ]
    )
    corner_points = [[0, 0], [size0, 0], [size0, size1], [0, size1]]
    return side_lines, corner_points


def num_side_pnt_changed(num_pnt, corner_points, side_lines):
    n_pnts_to_make = num_pnt - 2
    onside_data = corner_points.copy()
    if n_pnts_to_make > 0:
        for i in range(4):
            da_line = side_lines[i]
            for j in range(n_pnts_to_make):
                inline_point = da_line[0] + (da_line[1] - da_line[0]) / (
                    n_pnts_to_make + 1
                ) * (j + 1)
                onside_data.append([int(inline_point[0]), int(inline_point[1])])
    return onside_data


def match_sides_points(rect_atlas, size_atlas, rect_image, size_image):
    x_factor = rect_atlas[2] / rect_image[2]
    y_factor = rect_atlas[3] / rect_image[3]

    actual_right_dist_image = size_image[1] - (rect_image[0] + rect_image[2])
    actual_bottom_dist_image = size_image[0] - (rect_image[1] + rect_image[3])

    actual_right_dist_atlas = size_atlas[1] - (rect_atlas[0] + rect_atlas[2])
    actual_bottom_dist_atlas = size_atlas[0] - (rect_atlas[1] + rect_atlas[3])

    left_dist_atlas = rect_image[0] * x_factor
    right_dist_atlas = actual_right_dist_image * x_factor

    top_dist_atlas = rect_image[1] * y_factor
    bottom_dist_atlas = actual_bottom_dist_image * y_factor

    if left_dist_atlas <= rect_atlas[0]:
        atlas_corner_x = int(rect_atlas[0] - left_dist_atlas)
        image_corner_x = 0
    else:
        atlas_corner_x = 0
        left_dist_image = rect_atlas[0] / x_factor
        image_corner_x = int(rect_image[0] - left_dist_image)

    if right_dist_atlas <= actual_right_dist_atlas:
        atlas_right_side = size_atlas[1] - (actual_right_dist_atlas - right_dist_atlas)
        atlas_width = int(atlas_right_side - atlas_corner_x)
        image_width = size_image[1] - image_corner_x
    else:
        atlas_width = size_atlas[1] - atlas_corner_x
        right_dist_image = actual_right_dist_atlas / x_factor
        image_width = int(
            size_image[1]
            - image_corner_x
            - (actual_right_dist_image - right_dist_image)
        )

    if top_dist_atlas <= rect_atlas[1]:
        atlas_corner_y = int(rect_atlas[1] - top_dist_atlas)
        image_corner_y = 0
    else:
        atlas_corner_y = 0
        top_dist_image = rect_atlas[1] / y_factor
        image_corner_y = int(rect_image[1] - top_dist_image)

    if bottom_dist_atlas <= actual_bottom_dist_atlas:
        atlas_height = int(
            size_atlas[0]
            - atlas_corner_y
            - (actual_bottom_dist_atlas - bottom_dist_atlas)
        )
        image_height = size_image[0] - image_corner_y
    else:
        atlas_height = size_atlas[0] - atlas_corner_y
        bottom_dist_image = actual_bottom_dist_atlas / y_factor
        image_height = int(
            size_image[0]
            - image_corner_y
            - (actual_bottom_dist_image - bottom_dist_image)
        )

    atlas_rect = (atlas_corner_x, atlas_corner_y, atlas_width, atlas_height)
    image_rect = (image_corner_x, image_corner_y, image_width, image_height)

    return atlas_rect, image_rect


def get_corner_line_from_rect(rect):
    corners = [
        [rect[0], rect[1]],
        [rect[0] + rect[2] - 1, rect[1]],
        [rect[0] + rect[2] - 1, rect[1] + rect[3] - 1],
        [rect[0], rect[1] + rect[3] - 1],
    ]

    lines = np.asarray(
        [
            [corners[0], corners[1]],
            [corners[1], corners[2]],
            [corners[3], corners[2]],
            [corners[0], corners[3]],
        ]
    )

    return corners, lines


def make_label_rgb_img(label_img, lut):
    fimg = np.dstack([label_img, label_img, label_img])
    unique_label = np.unique(label_img).astype(int)
    for ind in np.unique(unique_label):
        if ind == 0:
            continue
        loc = np.where(label_img == ind)
        # print(lut[ind])
        fimg[loc[0], loc[1], 0] = lut[ind][0]
        fimg[loc[0], loc[1], 1] = lut[ind][1]
        fimg[loc[0], loc[1], 2] = lut[ind][2]
    return fimg.astype("uint8")


def make_contour_img(lable_img):
    unique_label = np.unique(lable_img).astype("int")
    img_size = lable_img.shape
    contour_img = np.zeros(img_size, "uint8")
    for label_ind in unique_label:
        if label_ind == 0:
            continue
        temp = np.zeros(img_size, "uint8")
        temp[lable_img == label_ind] = 1
        ct, hc = cv2.findContours(
            image=temp, mode=cv2.RETR_TREE, method=cv2.CHAIN_APPROX_NONE
        )
        for j in range(len(ct)):
            da_contour = ct[j].copy()
            da_shp = da_contour.shape
            da_contour = np.reshape(da_contour, (da_shp[0], da_shp[2]))
            contour_img[da_contour[:, 1], da_contour[:, 0]] = 1
    return contour_img


def get_tri_lines(rect, pnts):
    # print(rect)
    subdiv = cv2.Subdiv2D(rect)
    for p in pnts:
        # print(p)
        subdiv.insert((int(p[0]), int(p[1])))
    edge_list = subdiv.getEdgeList()
    lines_list = []
    # special_pnt = []
    for el in edge_list:
        pt1 = [el[0], el[1]]
        pt2 = [el[2], el[3]]
        if rect_contains(rect, pt1) and rect_contains(rect, pt2):
            lines_list.append(el)
        # else:
        #     if not rect_contains(rect, pt1) and pt1 not in special_pnt:
        #         special_pnt.append(pt1)
        #     if not rect_contains(rect, pt2) and pt2 not in special_pnt:
        #         special_pnt.append(pt2)
    return lines_list  # , special_pnt


# czi_img = image_file.data['scene 0'].copy()
# channel_hsv = image_file.hsv_colors
# temp_img = merge_channels_into_single_img(czi_img, channel_hsv)


def gamma_correction(src, gamma):
    inv_gamma = 1 / gamma

    table = [((i / 255) ** inv_gamma) * 255 for i in range(256)]
    table = np.array(table, np.uint8)

    return cv2.LUT(src, table)


def get_lower_val(val, tol, lim):
    lower_val = val - tol if tol < val else lim
    return lower_val


def get_upper_val(val, tol, lim):
    upper_val = val + tol
    upper_val = upper_val if upper_val <= lim else lim
    return upper_val


def get_bound_color(color, tol, level, mode):
    tol = float(tol)
    if mode == "gray":
        lower_val = get_lower_val(color, tol, 0)
        upper_val = get_upper_val(color, tol, level)
    elif mode == "hsv":
        lower_val = [get_lower_val(color[0], 1.0, 0)]
        upper_val = [get_upper_val(color[0], 1.0, 180)]
        for i in range(1, 3):
            lower_val.append(get_lower_val(color[i], tol, 0))
            upper_val.append(get_upper_val(color[i], tol, 255))
    else:
        lower_val = []
        upper_val = []
        for i in range(3):
            lower_val.append(get_lower_val(color[i], tol, 0))
            upper_val.append(get_upper_val(color[i], tol, level))
    return lower_val, upper_val


# ----------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------


def get_statusbar_style(col):
    style = (
        "background-color: #323232;"
        " color: {}; border-top: 1px solid #272727; "
        "padding-left: 30px;".format(col)
    )
    return style


def rotate(image, angle, img_center=None, scale=1.0):
    (h, w) = image.shape[:2]

    if img_center is None:
        img_center = (w // 2, h // 2)

    rot_mat = cv2.getRotationMatrix2D(img_center, angle, scale)
    rotated_img = cv2.warpAffine(image, rot_mat, (w, h))

    return rotated_img


def rotate_bound(image, angle):
    (h, w) = image.shape[:2]
    (center_x, center_y) = (w / 2, h / 2)

    rot_mat = cv2.getRotationMatrix2D((center_x, center_y), -angle, 1.0)
    cos = np.abs(rot_mat[0, 0])
    sin = np.abs(rot_mat[0, 1])

    bound_w = int((h * sin) + (w * cos))
    bound_h = int((h * cos) + (w * sin))

    rot_mat[0, 2] += (bound_w / 2) - center_x
    rot_mat[1, 2] += (bound_h / 2) - center_y

    return cv2.warpAffine(image, rot_mat, (bound_w, bound_h))


def center_resize(img, dim):
    img_shape = img.shape
    width, height = img_shape[1], img_shape[0]
    scale_factor = np.min(np.array([dim[0] / width, dim[1] / height]))
    resize_dim = (int(width * scale_factor), int(height * scale_factor))
    resize_img = cv2.resize(img, resize_dim, interpolation=cv2.INTER_LINEAR)

    y = int(0.5 * (dim[0] - resize_dim[0]))
    x = int(0.5 * (dim[1] - resize_dim[1]))

    if len(img_shape) == 3:
        center_img = np.zeros((dim[1], dim[0], img.shape[2])).astype(img.dtype)
    else:
        center_img = np.zeros((dim[1], dim[0])).astype(img.dtype)

    center_img[x : (x + resize_dim[1]), y : (y + resize_dim[0])] = resize_img

    return center_img


def get_tb_size(img_size):
    scale_factor = np.max(np.ravel(img_size) / 80)
    tb_size = (int(img_size[1] / scale_factor), int(img_size[0] / scale_factor))
    return tb_size


def get_slice_atlas_coord(points, cut, size, width, height, distance, origin):
    """

    :param points: pnts on image
    :param cut: the cut of slice image
    :param size: the size (height, width) of slice image in pixel
    :param width: width of image in mm
    :param height: height of image in mm
    :param distance: distance of slice with respect to Bregma in mm
    :param origin: the coord of bregma on the current slice image
    :return:
    """
    width_factor = width / size[1] * 1000
    height_factor = height / size[0] * 1000
    if cut == "Coronal":
        y_val = np.repeat(distance * 1000, len(points))
        x_val = (points[:, 0] - origin[0]) * width_factor
        z_val = (points[:, 1] - origin[1]) * height_factor
    elif cut == "Sagittal":
        x_val = np.repeat(distance * 1000, len(points))
        y_val = (points[:, 0] - origin[0]) * width_factor
        z_val = (points[:, 1] - origin[1]) * height_factor
    else:
        z_val = np.repeat(distance * 1000, len(points))
        y_val = (points[:, 0] - origin[0]) * width_factor
        x_val = (points[:, 1] - origin[1]) * height_factor
    return x_val, y_val, z_val


def delete_points_inside_eraser(points, ct, r):
    x_min, y_min, x_max, y_max = ct[0] - r, ct[1] - r, ct[0] + r, ct[1] + r
    x_bool = np.logical_and(points[:, 0] <= x_max, points[:, 0] >= x_min)
    y_bool = np.logical_and(points[:, 1] <= y_max, points[:, 1] >= y_min)
    chk_inds = np.where(np.logical_and(x_bool, y_bool))[0]
    if len(chk_inds) == 0:
        return None, None
    chk_pnt = points[chk_inds]
    dist = np.sum(np.power(chk_pnt - ct, 2), 1)
    del_ind = np.where(dist < np.power(r, 2))[0]
    if len(del_ind) == 0:
        return None, None
    real_del_ind = chk_inds[del_ind]
    remain_inds = np.zeros(len(points)) < 1
    remain_inds[real_del_ind] = False
    remain_points = points[remain_inds]
    return remain_points, real_del_ind


def interpolate_contour_points(points):
    if not np.all(points[-1] == points[0]):
        points[:, 0] = np.r_[points[:, 0], points[:, 0]]
        points[:, 1] = np.r_[points[:, 1], points[:, 1]]

    tck = splprep([points[:, 0], points[:, 1]], s=0, per=True)

    xi, yi = splev(np.linspace(0, 1, 1000), tck[0])
    return xi, yi


def create_vis_img(size, point_data, color, vis_type="p", closed=False):
    if isinstance(point_data, list):
        point_data = np.asarray(point_data).astype(int)
    else:
        point_data = point_data.astype(int)
    img = np.zeros((size[0], size[1], 3), "uint8")
    da_color = (int(color[0]), int(color[1]), int(color[2]))
    if vis_type == "p":
        for i in range(len(point_data)):
            cv2.circle(
                img,
                (int(point_data[i][0]), int(point_data[i][1])),
                radius=5,
                color=da_color,
                thickness=-1,
            )
    else:
        if closed:
            cv2.fillPoly(img, pts=[point_data], color=da_color)
        else:
            for i in range(len(point_data) - 1):
                cv2.line(
                    img,
                    (int(point_data[i][0]), int(point_data[i][1])),
                    (int(point_data[i + 1][0]), int(point_data[i + 1][1])),
                    color=da_color,
                    thickness=8,
                )
    return img


def color_vis_img(img, color):
    temp = np.zeros((img.shape[0], img.shape[1], 3), "uint8")
    temp[img != 0, 0] = color[0]
    temp[img != 0, 1] = color[1]
    temp[img != 0, 2] = color[2]
    return temp


def check_loaded_project(project_dict):
    all_keys = list(project_dict.keys())
    valid_keys = [
        "atlas_path",
        "current_atlas",
        "num_windows",
        "probe_type",
        "np_onside",
        "processing_slice",
        "processing_img",
        "overlay_img",
        "atlas_control",
        "img_ctrl_data",
        "setting_data",
        "tool_data",
        "layer_data",
        "working_img_data",
        "working_atlas_data",
        "object_data",
    ]
    valid_project = True
    for da_key in all_keys:
        if da_key not in valid_keys:
            valid_project = False
    return valid_project


def check_bounding_contains(points, size):
    if np.any(points[:, 0] > size[1]) or np.any(points[:, 1] > size[0]):
        return False
    else:
        return True


def obj_data_to_mesh3d(filename):
    vertices = []
    faces = []

    with open(filename, "r") as objf:
        for line in objf:
            slist = line.split()
            if slist:
                if slist[0] == "v":
                    vertex = np.array(slist[1:], dtype=float)
                    vertices.append(vertex)
                elif slist[0] == "f":
                    face = []
                    for k in range(1, len(slist)):
                        face.append(
                            [int(s) for s in slist[k].replace("//", "/").split("/")]
                        )
                    if len(face) > 3:  # triangulate the n-polyonal face, n>3
                        faces.extend(
                            [
                                [face[0][0] - 1, face[k][0] - 1, face[k + 1][0] - 1]
                                for k in range(1, len(face) - 1)
                            ]
                        )
                    else:
                        faces.append([face[j][0] - 1 for j in range(len(face))])
                else:
                    pass

    return np.array(vertices), np.array(faces)


def make_atlas_label_contour(atlas_folder, segmentation_data):
    sagital_contour_img = np.zeros(segmentation_data.shape, "i")
    coronal_contour_img = np.zeros(segmentation_data.shape, "i")
    horizontal_contour_img = np.zeros(segmentation_data.shape, "i")

    # pre-process boundary
    for i in range(len(segmentation_data)):
        da_slice = segmentation_data[i, :, :].copy()
        contour_img = make_contour_img(da_slice)
        sagital_contour_img[i, :, :] = contour_img

    for i in range(segmentation_data.shape[1]):
        da_slice = segmentation_data[:, i, :].copy()
        contour_img = make_contour_img(da_slice)
        coronal_contour_img[:, i, :] = contour_img

    for i in range(segmentation_data.shape[2]):
        da_slice = segmentation_data[:, :, i].copy()
        contour_img = make_contour_img(da_slice)
        horizontal_contour_img[:, :, i] = contour_img

    boundary = {
        "s_contour": sagital_contour_img,
        "c_contour": coronal_contour_img,
        "h_contour": horizontal_contour_img,
    }

    bnd = {"data": boundary}

    outfile_ct = open(os.path.join(atlas_folder, "contour_pre_made.pkl"), "wb")
    pickle.dump(bnd, outfile_ct)
    outfile_ct.close()

    return boundary


def get_angle_two_vector(vec1, vec2):
    """
    in 2d
    :param vec1:
    :param vec2:
    :return:
    """

    vec = np.array([-vec2[1], vec2[0]])
    b_coord = np.dot(vec1, vec2)
    p_coord = np.dot(vec1, vec)
    da_ang = np.arctan2(p_coord, b_coord)

    return da_ang


def rotate_base_points(data, base_loc):
    temp = np.stack([base_loc, np.repeat(0, len(base_loc))], axis=1)
    vec1 = data[1] - data[0]
    vec2 = np.array([0, 1])
    ang = get_angle_two_vector(vec1, vec2)
    rotm = np.array([[np.cos(ang), -np.sin(ang)], [np.sin(ang), np.cos(ang)]])
    base_pnt = np.dot(rotm, temp.T).T
    start_pnt = base_pnt + data[0]
    end_pnt = base_pnt + data[1]
    return start_pnt, end_pnt


def get_cell_count(cell_layer_index):
    cell_count = [0 for _ in range(5)]
    cell_layer_index = np.ravel(cell_layer_index)
    for i in range(5):
        cell_count[i] = np.sum(cell_layer_index == i)
    return cell_count


def load_point_data(data_file_path):
    file_basename = os.path.basename(data_file_path)
    file_name, file_ext = os.path.splitext(file_basename)
    msg = None
    data = None
    try:
        if file_ext == ".npy":
            data = np.load(data_file_path)
        elif file_ext == ".pkl":
            with open(data_file_path, "rb") as f:
                data = pickle.load(f)
            # data = data_file['points']
    except (
        IOError,
        OSError,
        ValueError,
        KeyError,
        IndexError,
        pickle.PickleError,
        pickle.UnpicklingError,
    ):
        msg = "Can not open atlas axis information file, please check the Tutorial on GitHub."
        return data, msg

    if not isinstance(data, np.ndarray):
        msg = "Data is not numpy ndarray, please check the Tutorial on GitHub."
        return data, msg

    data_shape = data.shape
    if len(data_shape) != 2:
        msg = "Data has wrong size, please check the Tutorial on GitHub."
        return data, msg

    if data_shape[1] != 3:
        msg = "Data has wrong size, please check the Tutorial on GitHub."
        return data, msg

    return data, msg



import os
from os.path import dirname, realpath, join
import sys

from .herbsgui import main


def run_herbs():
    current_wd = os.getcwd()
    script_dir = dirname(realpath(__file__))
    os.chdir(script_dir)
    main()
    os.chdir(current_wd)



import os
import sys
import numpy as np
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
from PyQt5.QtCore import *
from .uuuuuu import hsv2rgb, gamma_line, get_qhsv_from_czi_hsv, make_hist_data, read_qss_file


class BWSpin(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self)
        # wrap_frame = QFrame()
        wrap_layout = QVBoxLayout(self)
        wrap_layout.setContentsMargins(0, 0, 0, 0)
        wrap_layout.setSpacing(0)

        self.spin_nam = QLabel()
        self.spin_nam.setAlignment(Qt.AlignCenter)
        self.spin_val = QSpinBox()

        wrap_layout.addWidget(self.spin_nam)
        wrap_layout.addWidget(self.spin_val)


class GammaSpin(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self)
        # wrap_frame = QFrame()
        wrap_layout = QVBoxLayout(self)
        wrap_layout.setContentsMargins(0, 0, 0, 0)
        wrap_layout.setSpacing(0)

        self.spin_nam = QLabel()
        self.spin_nam.setAlignment(Qt.AlignCenter)
        self.spin_val = QDoubleSpinBox()
        self.spin_val.setDecimals(2)

        wrap_layout.addWidget(self.spin_nam)
        wrap_layout.addWidget(self.spin_val)


class ColorCombo(QComboBox):
    def __init__(self, parent=None):
        QComboBox.__init__(self)
        color_combo_style = read_qss_file('qss/color_combo.qss')
        self.setStyleSheet(color_combo_style)
        self.px = QPixmap(80, 30)
        n_vals = ['Light Blue', 'Blue', 'Green', 'Lime Green', 'Yellow', 'Olive', 'Red', 'Dark Red', 'Violet', 'Purple',
                  'Orange', 'Dark Orange', 'Turquoise', 'Blue Green']
        h_vals = [199, 235, 103, 115, 60, 60, 5, 4, 295, 279, 21, 20, 181, 151]
        s_vals = [60, 100, 70, 57, 67, 55, 85, 88, 77,  84, 81, 82, 51, 46]
        v_vals = [98, 96, 98, 59, 100, 60, 90, 54, 97,  77, 92, 73, 99, 79]
        s_vals = np.ravel(s_vals) * 0.01 * 255
        s_vals = s_vals.astype('uint8')
        v_vals = np.ravel(v_vals) * 0.01 * 255
        v_vals = v_vals.astype('uint8')
        self.hsv_color_list = []
        for i in range(len(h_vals)):
            self.hsv_color_list.append((h_vals[i] / 360., s_vals[i] / 255., v_vals[i] / 255.))
            self.px.fill(QColor.fromHsv(h_vals[i], s_vals[i], v_vals[i]))
            self.addItem(QIcon(self.px), n_vals[i])

        combo_list = QListView(self)
        self.setView(combo_list)


class ChannelSelector(QWidget):
    class SignalProxy(QObject):
        visChannels = pyqtSignal(object, object)
        changeColors = pyqtSignal(object, object)

    def __init__(self, rgb=False, parent=None):
        self._sigprox = ChannelSelector.SignalProxy()
        self.sig_vis_channels = self._sigprox.visChannels
        self.sig_change_color = self._sigprox.changeColors

        QWidget.__init__(self)

        channel_button_style = read_qss_file('qss/channel_selector.qss')
        self.vis = True
        self.setFixedSize(60, 60)
        self.setStyleSheet(channel_button_style)

        # wrap_frame = QFrame()
        wrap_layout = QVBoxLayout(self)
        wrap_layout.setContentsMargins(0, 0, 0, 0)
        wrap_layout.setSpacing(0)

        self.vis_btn = QPushButton()
        self.vis_btn.setFixedSize(60, 30)
        self.vis_btn.setCheckable(True)
        self.vis_btn.clicked.connect(self.change_vis)
        self.color_combo = ColorCombo()
        self.color_combo.setFixedSize(60, 28)
        self.color_combo.currentIndexChanged.connect(self.selection_change)
        self.color_label = QLabel()

        self.color_label.setVisible(False)

        wrap_layout.addWidget(self.vis_btn)
        wrap_layout.addWidget(self.color_combo)
        wrap_layout.addWidget(self.color_label)

    def set_channel_index(self, index):
        self.index = index

    def change_vis(self):
        if self.vis_btn.isChecked():
            self.set_checked(True)
        else:
            self.set_checked(False)
        self.sig_vis_channels.emit(self.vis, self.index)

    def set_checked(self, is_checked):
        if is_checked:
            self.color_label.setVisible(True)
            self.color_combo.setVisible(False)
            self.vis = False
        else:
            self.color_label.setVisible(False)
            self.color_combo.setVisible(True)
            self.vis = True

    def selection_change(self):
        da_index = self.color_combo.currentIndex()
        da_color = self.color_combo.hsv_color_list[da_index]
        self.sig_change_color.emit(da_color, self.index)

    def add_item(self, hsv_color):
        self.color_combo.hsv_color_list.append((hsv_color[0], hsv_color[1], hsv_color[2]))
        q_hsv = get_qhsv_from_czi_hsv(hsv_color)
        self.color_combo.px.fill(QColor.fromHsv(q_hsv[0], q_hsv[1], q_hsv[2]))
        self.color_combo.addItem(QIcon(self.color_combo.px), '')
        self.color_combo.setCurrentIndex(len(self.color_combo.hsv_color_list) - 1)

    def delete_item(self):
        self.color_combo.removeItem(len(self.color_combo.hsv_color_list) - 1)
        self.color_combo.hsv_color_list.pop()




"""
``HERBS Code``
================

Functions present in herbs are listed below.


For Image Processing
--------------------------

   ...
   ...

For Atlas
--------------

   ...

For Others
------------------

   ...


"""

from .run_herbs import run_herbs
from .czi_reader import CZIReader


import os
import sys
import numpy as np
import pyqtgraph as pg
import pyqtgraph.functions as fn
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
from PyQt5.QtCore import *

import cv2

from .wtiles import *


btm_style = '''

QPushButton {
    border : None; 
    background: transparent;
    margin: 0px;
    padding-top: 0px;
    border-radius: 0px;
    min-width: 24px;
    min-height: 24px;
}
QPushButton:checked {
    background-color: #383838; 
    border: 1px solid #636363; 
}

QPushButton:pressed {
    background-color: #383838; 
    border: 1px solid #636363; 
}

QPushButton:hover {
    background-color: #383838; 
    border: 1px solid #636363; 
}

'''


eye_button_style = '''
    border-left: None;
    border-right: 1px solid rgb(128, 128, 128);
    border-top: None;
    border-bottom: None;
    border-radius:0px;
    background: transparent;
'''

text_btn_style = '''
QDoubleButton {
    background: transparent;
    border: None;
    border-radius:0px;
    text-align:left; 
    padding-left: 5px; 
    padding-right: 0px;
    color:rgb(240, 240, 240);
    margin: 0px;
    height: 60px;
    width: 160px;
}

QDoubleButton:disabled {
    color: rgb(190, 190, 190);
}
'''

line_edit_style = '''
QLineEdit {
    background: transparent;
    border: 1px solid green;
    border-radius:0px;
    text-align:left; 
    color: white; 
    padding-left: 5px;    
    margin: 0px;
    height: 60px;
    width: 160px; 
}
'''


class SingleLayer(QWidget):
    sig_clicked = pyqtSignal(object)
    eye_clicked = pyqtSignal(object)
    sig_delete = pyqtSignal(object)

    def __init__(self, parent=None, layer_id=0, link='None'):
        QWidget.__init__(self, parent=parent)

        self.inactive_style = 'QFrame{background-color:rgb(83, 83, 83); border: 1px solid rgb(128, 128, 128);}'
        self.active_style = 'QFrame{background-color:rgb(107, 107, 107); border: 1px solid rgb(128, 128, 128);}'

        # self.setFixedWidth(280)
        self.id = layer_id
        self.link = link
        self.active = True
        self.vis = True
        self.thumbnail_data = None
        self.eye_button = QPushButton()
        self.eye_button.setFixedSize(QSize(30, 60))
        self.eye_button.setStyleSheet(eye_button_style)
        self.eye_button.setCheckable(True)
        eye_icon = QIcon()
        eye_icon.addPixmap(QPixmap("icons/layers/eye_on.png"), QIcon.Normal, QIcon.Off)
        eye_icon.addPixmap(QPixmap("icons/layers/eye_off.png"), QIcon.Normal, QIcon.On)
        self.eye_button.setIcon(eye_icon)
        self.eye_button.setIconSize(QSize(20, 20))
        self.eye_button.clicked.connect(self.eye_on_click)

        self.tbnail = QPushButton()
        self.tbnail.setStyleSheet('border:1px solid black; border-radius:0px;')
        self.tbnail.setFixedWidth(50)
        self.tbnail.clicked.connect(self.on_click)

        self.text_btn = QPushButton()
        self.text_btn.setStyleSheet(text_btn_style)
        self.text_btn.setText(link)
        self.text_btn.setFixedSize(QSize(140, 60))
        self.text_btn.clicked.connect(self.on_click)
        # self.text_btn.double_clicked.connect(self.on_doubleclick)

        # self.l_line_edit = QLineEdit()
        # self.l_line_edit.setStyleSheet(line_edit_style)
        # self.l_line_edit.setFixedSize(QSize(140, 60))
        # self.l_line_edit.editingFinished.connect(self.enter_pressed)
        # self.l_line_edit.setVisible(False)

        self.trash_button = QPushButton()
        self.trash_button.setFixedSize(QSize(25, 60))
        self.trash_button.setStyleSheet(eye_button_style)
        self.trash_button.setIcon(QIcon('icons/sidebar/trash.png'))
        self.trash_button.setIconSize(QSize(20, 20))
        self.trash_button.clicked.connect(self.on_delete)

        self.inner_frame = QFrame()
        self.inner_frame.setStyleSheet(self.active_style)
        self.inner_layout = QHBoxLayout(self.inner_frame)
        self.inner_layout.setContentsMargins(0, 0, 0, 0)
        self.inner_layout.setSpacing(0)
        self.inner_layout.setAlignment(Qt.AlignVCenter)
        self.inner_layout.addWidget(self.eye_button)
        self.inner_layout.addSpacing(5)
        self.inner_layout.addWidget(self.tbnail)
        self.inner_layout.addSpacing(5)
        self.inner_layout.addWidget(self.text_btn)
        # self.inner_layout.addWidget(self.l_line_edit)
        self.inner_layout.addWidget(self.trash_button)
        self.inner_layout.addStretch()

        outer_layout = QHBoxLayout()
        outer_layout.setContentsMargins(0, 0, 0, 0)
        outer_layout.setSpacing(0)
        outer_layout.setAlignment(Qt.AlignVCenter)
        outer_layout.addWidget(self.inner_frame)

        self.setLayout(outer_layout)
        self.setFixedHeight(60)

    def set_checked(self, check):
        self.active = check
        if not self.active:
            self.inner_frame.setStyleSheet(self.inactive_style)
        else:
            self.inner_frame.setStyleSheet(self.active_style)

    def on_click(self):
        self.set_checked(True)
        self.sig_clicked.emit(self.id)

    def eye_on_click(self):
        if self.eye_button.isChecked():
            self.vis = False
        else:
            self.vis = True
        self.tbnail.setEnabled(self.vis)
        self.text_btn.setEnabled(self.vis)
        # self.l_line_edit.setEnabled(self.vis)
        self.eye_clicked.emit((self.id, self.link, self.vis))

    def on_delete(self):
        self.sig_delete.emit(self.id)

    # @pyqtSlot()
    # def on_doubleclick(self):
    #     self.l_line_edit.setText(self.text_btn.text())
    #     self.l_line_edit.setVisible(True)
    #     self.text_btn.setVisible(False)
    #     self.l_line_edit.setFocus(True)
    #     self.on_click()
    #
    # def enter_pressed(self):
    #     self.l_line_edit.setVisible(False)
    #     self.text_btn.setText(self.l_line_edit.text())
    #     self.text_btn.setVisible(True)

    def set_thumbnail_data(self, thumbnail_data):
        if len(thumbnail_data.shape) == 2:
            im = np.zeros((thumbnail_data.shape[0], thumbnail_data.shape[1], 3)).astype(np.uint8)
            # im = thumbnail_data.copy()
            im[:, :, 0] = thumbnail_data
            im[:, :, 1] = thumbnail_data
            im[:, :, 2] = thumbnail_data
        else:
            im = thumbnail_data.astype('uint8')

        self.thumbnail_data = im.copy()

        im_shape = np.ravel(im.shape[:2])

        size = QSize(int(im_shape[1] / 2), int(im_shape[0] / 2))
        image = QImage(im.data, im_shape[1], im_shape[0], im.strides[0], QImage.Format_RGB888)
        pixmap = QPixmap(image)

        self.tbnail.setIcon(QIcon(pixmap.scaled(size)))
        self.tbnail.setIconSize(size)

    def is_checked(self):
        return self.active

    def get_link(self):
        return self.link


class LayersControl(QWidget):

    class SignalProxy(QObject):
        sigOpacityChanged = pyqtSignal(object)
        sigVisChanged = pyqtSignal(object)
        sigBlendModeChanged = pyqtSignal(object)
        sigDeleteLayer = pyqtSignal(object)

    def __init__(self, parent=None):

        self._sigprox = LayersControl.SignalProxy()
        self.sig_opacity_changed = self._sigprox.sigOpacityChanged
        self.sig_visible_changed = self._sigprox.sigVisChanged
        self.sig_blend_mode_changed = self._sigprox.sigBlendModeChanged
        self.sig_layer_deleted = self._sigprox.sigDeleteLayer

        QWidget.__init__(self)

        self.valid_blendable_link = ['img-mask', 'img-virus', 'img-contour', 'img-overlay', 'Image', 'atlas-overlay']

        self.current_layer_index = []

        self.layer_list = []
        self.layer_id = []
        self.layer_link = []
        self.layer_opacity = []
        self.layer_blend_mode = []
        self.layer_color = []

        self.layer_count = 0

        blend_frame = QFrame()
        blend_layout = QHBoxLayout(blend_frame)
        blend_layout.setContentsMargins(0, 0, 0, 0)
        blend_layout.setSpacing(5)

        combo_label = QLabel('Composition:')
        self.layer_blend_combo = QComboBox()
        self.layer_blend_combo.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.layer_blend_combo.setEditable(False)
        combo_value = ['Plus', 'Multiply', 'Overlay', 'SourceOver']
        self.layer_blend_combo.addItems(combo_value)
        self.layer_blend_combo.setCurrentText('Plus')
        self.layer_blend_combo.currentTextChanged.connect(self.blend_mode_changed)

        blend_layout.addWidget(combo_label)
        blend_layout.addWidget(self.layer_blend_combo)

        # opacity wrap
        opacity_frame = QFrame()
        opacity_layout = QHBoxLayout(opacity_frame)
        opacity_layout.setContentsMargins(0, 0, 0, 0)
        opacity_layout.setSpacing(5)

        layer_opacity_label = QLabel('Opacity:')
        self.layer_opacity_slider = QSlider(Qt.Horizontal)
        # self.layer_opacity_slider.setFixedWidth(100)
        self.layer_opacity_slider.setMaximum(100)
        self.layer_opacity_slider.setMinimum(0)
        self.layer_opacity_slider.setValue(100)
        self.layer_opacity_slider.sliderMoved.connect(self.opacity_slider_moved)
        self.layer_opacity_slider.valueChanged.connect(self.change_opacity_label_value)
        self.layer_opacity_val_label = QLabel('100%')

        opacity_layout.addWidget(layer_opacity_label)
        opacity_layout.addWidget(self.layer_opacity_slider)
        opacity_layout.addWidget(self.layer_opacity_val_label)

        self.layer_frame = QFrame()
        self.layer_frame.setStyleSheet('background: transparent; border: 0px;')
        self.layer_frame.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.layer_layout = QBoxLayout(QBoxLayout.BottomToTop, self.layer_frame)
        self.layer_layout.setAlignment(Qt.AlignBottom)
        self.layer_layout.setContentsMargins(0, 0, 0, 0)
        self.layer_layout.setSpacing(0)

        self.layer_scroll = QScrollArea()
        self.layer_scroll.setStyleSheet('background: transparent; border: 0px;')
        self.layer_scroll.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.layer_scroll.setWidget(self.layer_frame)
        self.layer_scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.layer_scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.layer_scroll.setWidgetResizable(True)

        mid_frame = QFrame()
        mid_frame.setStyleSheet('background: transparent; border: 1px solid rgb(128, 128, 128);')
        mid_frame.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        mid_layout = QGridLayout(mid_frame)
        mid_layout.setContentsMargins(0, 0, 0, 0)
        mid_layout.setSpacing(0)
        mid_layout.setAlignment(Qt.AlignBottom)
        mid_layout.addWidget(self.layer_scroll, 0, 0, 1, 1)

        # extra btn
        self.add_layer_btn = QPushButton()
        self.add_layer_btn.setFixedSize(24, 24)
        self.add_layer_btn.setStyleSheet(btm_style)
        self.add_layer_btn.setIcon(QIcon('icons/layers/add.png'))
        self.add_layer_btn.setIconSize(QSize(20, 20))
        self.add_layer_btn.clicked.connect(lambda: self.add_layer('Layer'))

        self.delete_layer_btn = QPushButton()
        self.delete_layer_btn.setFixedSize(24, 24)
        self.delete_layer_btn.setStyleSheet(btm_style)
        self.delete_layer_btn.setIcon(QIcon('icons/layers/trash.png'))
        self.delete_layer_btn.setIconSize(QSize(20, 20))
        self.delete_layer_btn.clicked.connect(self.delete_layer_btn_clicked)

        outer_layout = QVBoxLayout()
        outer_layout.setSpacing(10)
        outer_layout.addWidget(blend_frame)
        outer_layout.addWidget(opacity_frame)
        outer_layout.addWidget(mid_frame)

        self.setLayout(outer_layout)

    def add_layer(self, widget_link, color):
        self.layer_id.append(self.layer_count)
        self.layer_link.append(widget_link)
        print(self.layer_link)
        self.layer_opacity.append(100)
        self.layer_blend_mode.append('Plus')
        self.sig_blend_mode_changed.emit((widget_link, 
                                          self.layer_blend_mode[-1]))
        self.layer_color.append(color)
        # self.layer_data.append(data)

        new_layer = SingleLayer(layer_id=self.layer_count, link=widget_link)
        new_layer.sig_clicked.connect(self.layer_clicked)
        new_layer.eye_clicked.connect(self.layer_eye_clicked)
        new_layer.sig_delete.connect(self.delete_layer_btn_clicked)

        active_index = len(self.layer_id) - 1
        print('add', active_index)
        if self.current_layer_index:
            for da_ind in self.current_layer_index:
                self.layer_list[da_ind].set_checked(False)
        self.current_layer_index = [active_index]

        self.layer_list.append(new_layer)
        self.layer_layout.addWidget(self.layer_list[-1])
        self.layer_count += 1

        if self.layer_opacity[-1] != self.layer_opacity_slider.value():
            self.layer_opacity_slider.setValue(self.layer_opacity[-1])
        if self.layer_blend_mode[-1] != self.layer_blend_combo.currentText():
            self.layer_blend_combo.setCurrentText(self.layer_blend_mode[-1])

    def master_layers(self, res, layer_type, color):
        if layer_type not in self.layer_link:
            self.add_layer(layer_type, color)
            self.layer_list[-1].set_thumbnail_data(res)
        else:
            da_ind = np.where(np.ravel(self.layer_link) == layer_type)[0][0]
            self.layer_list[da_ind].set_thumbnail_data(res)

    def delete_layer_btn_clicked(self, clicked_id):
        delete_index = np.where(np.ravel(self.layer_id) == clicked_id)[0][0]
        self.sig_layer_deleted.emit(self.layer_link[delete_index])
        self.delete_layer(delete_index)

    def delete_layer(self, delete_index):
        self.layer_layout.removeWidget(self.layer_list[delete_index])
        self.layer_list[delete_index].deleteLater()
        del self.layer_list[delete_index]
        del self.layer_id[delete_index]
        del self.layer_link[delete_index]
        del self.layer_opacity[delete_index]
        del self.layer_blend_mode[delete_index]
        del self.layer_color[delete_index]
        # del self.layer_data[delete_index]
        self.correct_current_layer_index(delete_index)

    def correct_current_layer_index(self, delete_index):
        if delete_index in self.current_layer_index:
            remove_ind = np.where(np.ravel(self.current_layer_index) == delete_index)[0][0]
            del self.current_layer_index[remove_ind]
        for i in range(len(self.current_layer_index)):
            if self.current_layer_index[i] > delete_index:
                self.current_layer_index[i] -= 1

    def layer_clicked(self, clicked_id):
        clicked_index = np.where(np.ravel(self.layer_id) == clicked_id)[0][0]

        modifiers = QApplication.keyboardModifiers()
        if modifiers == Qt.ControlModifier:
            if clicked_index in self.current_layer_index:
                self.layer_list[clicked_index].set_checked(False)
                remove_ind = np.where(np.ravel(self.current_layer_index) == clicked_index)[0][0]
                self.current_layer_index.pop(remove_ind)
            else:
                self.layer_list[clicked_index].set_checked(True)
                self.current_layer_index.append(clicked_index)
        else:
            for da_ind in self.current_layer_index:
                self.layer_list[da_ind].set_checked(False)
            self.layer_list[clicked_index].set_checked(True)
            self.current_layer_index = [clicked_index]

            if self.layer_opacity[clicked_index] != self.layer_opacity_slider.value():
                self.layer_opacity_slider.setValue(self.layer_opacity[clicked_index])
            if self.layer_blend_mode[clicked_index] != self.layer_blend_combo.currentText():
                self.layer_blend_combo.setCurrentText(self.layer_blend_mode[clicked_index])

    def layer_eye_clicked(self, event):
        self.sig_visible_changed.emit(event)

    def blend_mode_changed(self):
        if not self.current_layer_index:
            return
        if len(self.current_layer_index) > 1:
            return
        blend_mode = self.layer_blend_combo.currentText()
        if blend_mode != self.layer_blend_mode[self.current_layer_index[0]]:
            current_link = self.layer_link[self.current_layer_index[0]]
            self.layer_blend_mode[self.current_layer_index[0]] = blend_mode
            self.sig_blend_mode_changed.emit((current_link, blend_mode))

    def opacity_slider_moved(self):
        if not self.current_layer_index:
            return
        if len(self.current_layer_index) > 1:
            return
        val = self.layer_opacity_slider.value()
        if val != self.layer_opacity[self.current_layer_index[0]]:
            current_link = self.layer_link[self.current_layer_index[0]]
            current_color = self.layer_color[self.current_layer_index[0]]
            self.layer_opacity[self.current_layer_index[0]] = val
            self.sig_opacity_changed.emit((current_link, current_color, val))

    def change_opacity_label_value(self):
        da_val = self.layer_opacity_slider.value()
        self.layer_opacity_val_label.setText('{} %'.format(da_val))

    def get_layer_data(self):
        tb_list = []
        for i in range(len(self.layer_link)):
            da_tb = self.layer_list[i].thumbnail_data
            tb_list.append(da_tb)
        print(self.layer_link)
        data = {'layer_link': self.layer_link,
                'layer_color': self.layer_color,
                'layer_thumbnail': tb_list,
                'layer_opacity': self.layer_opacity,
                'layer_blend_mode': self.layer_blend_mode,
                'current_layer_index': self.current_layer_index}
        return data

    def set_layer_data(self, data):
        for i in range(len(data['layer_link'])):
            self.master_layers(data['layer_thumbnail'][i], data['layer_link'][i], data['layer_color'][i])
        self.layer_color = data['layer_color']
        self.layer_opacity = data['layer_opacity']
        self.layer_blend_mode = data['layer_blend_mode']
        self.current_layer_index = data['current_layer_index']
        self.layer_list[-1].set_checked(False)
        for i in range(len(self.current_layer_index)):
            self.layer_list[i].set_checked(True)

    def clear_all(self):
        ind = np.arange(len(self.layer_list))[::-1]
        for i in ind:
            self.layer_layout.removeWidget(self.layer_list[i])
            self.layer_list[i].deleteLater()
            del self.layer_list[i]
        self.layer_link = []
        self.layer_id = []
        self.layer_link = []
        self.layer_color = []
        self.layer_opacity = []
        self.layer_blend_mode = []
        self.layer_count = []
        self.current_layer_index = []
        self.layer_count = 0














from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *
import numpy as np
from .uuuuuu import read_qss_file

dialog_style = '''
QDialog {
    background-color: rgb(50, 50, 50);
    color: white;
    font-size: 12px;
    width: 50px;
}

QLabel {
    color: white;
    width: 50px;
}
'''


class LayerSettingDialog(QDialog):
    def __init__(self, window_name, min_val, max_val, val):
        super().__init__()

        self.setWindowTitle(window_name)
        self.setStyleSheet(dialog_style)

        self.val = val

        self.val_slider = QSlider(Qt.Horizontal)
        self.val_slider.setValue(val)
        self.val_slider.setMinimum(min_val)
        self.val_slider.setMaximum(max_val)
        self.val_slider.setSingleStep(1)
        self.val_slider.sliderMoved.connect(self.val_spinbox_changed)

        self.val_spinbox = QSpinBox()
        self.val_spinbox.setValue(val)
        self.val_spinbox.setMaximum(max_val)
        self.val_spinbox.setMinimum(min_val)
        self.val_spinbox.setSingleStep(1)
        self.val_spinbox.valueChanged.connect(self.value_changed)

        # ok button, used to close window
        ok_btn = QDialogButtonBox(QDialogButtonBox.Ok)
        ok_btn.accepted.connect(self.accept)

        # add widget to layout
        layout = QHBoxLayout()
        layout.addWidget(self.val_slider)
        layout.addWidget(self.val_spinbox)
        layout.addSpacing(10)
        layout.addWidget(ok_btn)
        self.setLayout(layout)

    def accept(self) -> None:
        self.close()

    def val_spinbox_changed(self):
        val = self.val_slider.value()
        self.val = val
        self.val_spinbox.setValue(val)

    def value_changed(self):
        val = self.val_spinbox.value()
        self.val = val
        self.val_slider.setValue(val)


class SliceSettingDialog(QDialog):
    def __init__(self):
        super().__init__()

        self.setWindowTitle('Current Slice Settings')
        self.setStyleSheet(dialog_style)

        self.cut = 'Coronal'
        self.width = 0
        self.height = 0
        self.distance = 0

        self.cut_combo = QComboBox()
        self.cut_combo.addItems(['Coronal', 'Sagittal', 'Horizontal'])
        self.cut_combo.currentIndexChanged.connect(self.cut_changed)

        width_label = QLabel('Width (mm):')
        height_label = QLabel('Height (mm):')
        distance_label = QLabel('Distance w.r.t. Bregma (mm):')

        self.width_val = QDoubleSpinBox()
        self.width_val.setValue(0)
        self.width_val.setRange(-50, 50)
        self.width_val.valueChanged.connect(self.width_val_changed)
        self.height_val = QDoubleSpinBox()
        self.height_val.setValue(0)
        self.height_val.setRange(-50, 50)
        self.height_val.valueChanged.connect(self.height_val_changed)
        self.distance_val = QDoubleSpinBox()
        self.distance_val.setValue(0)
        self.distance_val.setRange(-50, 50)
        self.distance_val.valueChanged.connect(self.distance_val_changed)

        # ok button, used to close window
        ok_btn = QDialogButtonBox(QDialogButtonBox.Ok)
        ok_btn.accepted.connect(self.accept)

        # add widget to layout
        content_frame = QFrame()
        content_layout = QGridLayout(content_frame)
        content_layout.addWidget(self.cut_combo, 0, 0, 1, 2)
        content_layout.addWidget(width_label, 1, 0, 1, 1)
        content_layout.addWidget(self.width_val, 1, 1, 1, 1)
        content_layout.addWidget(height_label, 2, 0, 1, 1)
        content_layout.addWidget(self.height_val, 2, 1, 1, 1)
        content_layout.addWidget(distance_label, 3, 0, 1, 1)
        content_layout.addWidget(self.distance_val, 3, 1, 1, 1)

        layout = QVBoxLayout()
        layout.addWidget(content_frame)
        layout.addWidget(ok_btn)
        self.setLayout(layout)

    def accept(self) -> None:
        self.close()

    def cut_changed(self):
        self.cut = self.cut_combo.currentText()

    def width_val_changed(self):
        self.width = self.width_val.value()

    def height_val_changed(self):
        self.height = self.height_val.value()

    def distance_val_changed(self):
        self.distance = self.distance_val.value()


class QDoubleButton(QPushButton):
    right_clicked = pyqtSignal()
    left_clicked = pyqtSignal()
    double_clicked = pyqtSignal()

    def __init__(self, *args, **kwargs):
        super(QDoubleButton, self).__init__(*args, **kwargs)

        btn_style = read_qss_file('qss/object_text_button.qss')
        self.setStyleSheet(btn_style)

        self.timer = QTimer()
        self.timer.setSingleShot(True)
        self.timer.setInterval(150)
        self.timer.timeout.connect(self.timeout)

        self.is_double = False
        self.is_left_click = True

        self.installEventFilter(self)

    def eventFilter(self, obj, event):
        if event.type() == QEvent.MouseButtonPress:
            if not self.timer.isActive():
                self.timer.start()

            self.is_left_click = False
            if event.button() == Qt.LeftButton:
                self.is_left_click = True

            return True

        elif event.type() == QEvent.MouseButtonDblClick:
            self.is_double = True
            return True

        return False

    def timeout(self):
        if self.is_double:
            self.double_clicked.emit()
        else:
            if self.is_left_click:
                self.left_clicked.emit()
            else:
                self.right_clicked.emit()

        self.is_double = False


class QLabelClickable(QLabel):
    # doubleClicked = pyqtSignal(object)
    clicked = pyqtSignal(object)

    def __init__(self, parent=None):
        super(QLabelClickable, self).__init__(parent)

    def mousePressEvent(self, event):
        self.ultimo = 'Click'

    def mouseReleaseEvent(self, event):
        if self.ultimo == 'Click':
            QTimer.singleShot(QApplication.instance().doubleClickInterval(),
                              self.performSingleClickAction)
        else:
            self.ultimo = 'doubleClick'
            self.clicked.emit(self.ultimo)

    def mouseDoubleClickEvent(self, event):
        self.ultimo = 'doubleClick'

    def performSingleClickAction(self):
        if self.ultimo == 'Click':
            self.clicked.emit(self.ultimo)
            self.update()


class QFrameClickable(QFrame):
    clicked = pyqtSignal(object)

    def __init__(self, parent=None):
        super(QFrameClickable, self).__init__(parent)

    # def mousePressEvent(self, QMouseEvent):
    #     if QMouseEvent.button() == Qt.LeftButton:
    #         self.clicked.emit()
    #     elif QMouseEvent.button() == Qt.RightButton:
    #         self.clicked.emit()
    #     self.update()

    def mousePressEvent(self, event):
        self.ultimo = 'Click'

    def mouseReleaseEvent(self, event):
        if self.ultimo == 'Click':
            QTimer.singleShot(QApplication.instance().doubleClickInterval(),
                              self.performSingleClickAction)
        else:
            self.ultimo = 'doubleClick'
            self.clicked.emit(self.ultimo)

    def mouseDoubleClickEvent(self, event):
        self.ultimo = 'doubleClick'

    def performSingleClickAction(self):
        if self.ultimo == 'Click':
            self.clicked.emit(self.ultimo)
            self.update()


class LineEditEntered(QLineEdit):
    sig_return_pressed = pyqtSignal()

    def __init__(self):
        QLineEdit.__init__(self)

    def keyPressEvent(self, event):
        super(LineEditEntered, self).keyPressEvent(event)

        if event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
            self.sig_return_pressed.emit()


class IntLineEdit(QLineEdit):
    sig_text_changed = pyqtSignal(object)
    def __init__(self, w_id, text=None, parent=None):
        QLineEdit.__init__(self)
        self.id = w_id
        if text is not None:
            self.setText(text)
        self.setValidator(QIntValidator())
        self.textChanged.connect(self.editing_text)

    def editing_text(self, text_val):
        self.sig_text_changed.emit((self.id, text_val))


class FaceCombo(QComboBox):
    sig_text_changed = pyqtSignal(object)
    def __init__(self, w_id, parent=None):
        QComboBox.__init__(self)
        self.id = w_id
        self.currentIndexChanged.connect(self.index_changed)

    def index_changed(self, index):
        self.sig_text_changed.emit((self.id, index))


class LinearSiliconInfoDialog(QDialog):
    def __init__(self, pss):
        super().__init__()

        self.setWindowTitle('Linear Silicon Probe Geometry Setting Window')
        self.setStyleSheet(dialog_style)

        btn_box = QDialogButtonBox.Ok | QDialogButtonBox.Cancel

        self.button_box = QDialogButtonBox(btn_box)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        if pss is None:
            self.probe_settings = {'probe_length': 0,
                                   'probe_thickness': 0,
                                   'tip_length': 0,
                                   'site_height': 0,
                                   'site_width': 0,
                                   'per_max_sites': [0],
                                   'sites_distance': [0],
                                   'x_bias': [0],
                                   'y_bias': [0]}
        else:
            self.probe_settings = pss

        n_column = len(self.probe_settings['x_bias'])

        fig_label = QLabel()
        pixmap = QPixmap('icons/toolbar/linear_silicon.png')
        fig_label.setPixmap(pixmap)
        fig_label.setFixedHeight(300)

        right_frame = QFrame()
        self.right_layout = QGridLayout(right_frame)

        probe_length_label = QLabel('A Probe Length (um): ')
        self.probe_length_input = IntLineEdit(0, str(self.probe_settings['probe_length']))
        self.probe_length_input.sig_text_changed.connect(self.probe_length_changed)

        thickness_label = QLabel('Probe Thickness (um): ')
        self.thickness_input = IntLineEdit(0, str(self.probe_settings['probe_thickness']))
        self.thickness_input.sig_text_changed.connect(self.probe_thickness_changed)

        tip_length_label = QLabel('B Tip Length (um): ')
        self.tip_length_input = IntLineEdit(0, str(self.probe_settings['tip_length']))
        self.tip_length_input.sig_text_changed.connect(self.tip_length_changed)

        site_height_label = QLabel('Site Height (um): ')
        self.site_height_input = IntLineEdit(0, str(self.probe_settings['site_height']))
        self.site_height_input.sig_text_changed.connect(self.site_height_changed)

        site_width_label = QLabel('Site Width (um): ')
        self.site_width_input = IntLineEdit(0, str(self.probe_settings['site_width']))
        self.site_width_input.sig_text_changed.connect(self.site_width_changed)

        n_column_label = QLabel('Number of columns: ')
        self.n_column_spinbox = QSpinBox()
        self.n_column_spinbox.setRange(1, 10)
        self.n_column_spinbox.setValue(n_column)

        row_names = [QLabel('Number of Sites: '), QLabel('C Sites Distance (um): '),
                     QLabel('D X Bias (um): '), QLabel('E Y Bias (um): ')]

        self.sites_distance_wl = []
        self.per_max_sites_wl = []
        self.x_bias_wl = []
        self.y_bias_wl = []
        for i in range(n_column):
            self.sites_distance_wl.append(IntLineEdit(i, str(self.probe_settings['sites_distance'][i])))
            self.per_max_sites_wl.append(IntLineEdit(i, str(self.probe_settings['per_max_sites'][i])))
            self.x_bias_wl.append(IntLineEdit(i, str(self.probe_settings['x_bias'][i])))
            self.y_bias_wl.append(IntLineEdit(i, str(self.probe_settings['y_bias'][i])))
            self.sites_distance_wl[-1].sig_text_changed.connect(self.sites_distance_changed)
            self.per_max_sites_wl[-1].sig_text_changed.connect(self.per_max_sites_changed)
            self.x_bias_wl[-1].sig_text_changed.connect(self.x_bias_changed)
            self.y_bias_wl[-1].sig_text_changed.connect(self.y_bias_changed)

        # connect
        self.n_column_spinbox.valueChanged.connect(self.n_column_changed)

        self.right_layout.addWidget(probe_length_label, 0, 0, 1, 1)
        self.right_layout.addWidget(self.probe_length_input, 0, 1, 1, 1)
        self.right_layout.addWidget(thickness_label, 1, 0, 1, 1)
        self.right_layout.addWidget(self.thickness_input, 1, 1, 1, 1)
        self.right_layout.addWidget(tip_length_label, 2, 0, 1, 1)
        self.right_layout.addWidget(self.tip_length_input, 2, 1, 1, 1)
        self.right_layout.addWidget(site_height_label, 3, 0, 1, 1)
        self.right_layout.addWidget(self.site_height_input, 3, 1, 1, 1)
        self.right_layout.addWidget(site_width_label, 4, 0, 1, 1)
        self.right_layout.addWidget(self.site_width_input, 4, 1, 1, 1)

        self.right_layout.addWidget(n_column_label, 5, 0, 1, 1)
        self.right_layout.addWidget(self.n_column_spinbox, 5, 1, 1, 1)

        for i in range(len(row_names)):
            self.right_layout.addWidget(row_names[i], 6 + i, 0, 1, 1)

        for j in range(n_column):
            self.right_layout.addWidget(self.per_max_sites_wl[j], 6, j + 1, 1, 1)
            self.right_layout.addWidget(self.sites_distance_wl[j], 7, j + 1, 1, 1)
            self.right_layout.addWidget(self.x_bias_wl[j], 8, j + 1, 1, 1)
            self.right_layout.addWidget(self.y_bias_wl[j], 9, j + 1, 1, 1)

        out_frame = QFrame()
        out_layout = QHBoxLayout(out_frame)
        out_layout.addWidget(fig_label)
        out_layout.addWidget(right_frame)

        # add widget to layout
        layout = QVBoxLayout()
        layout.addWidget(out_frame)
        layout.addSpacing(10)
        layout.addWidget(self.button_box)
        self.setLayout(layout)

    def n_column_changed(self):
        print('gjgjhgjhg')
        n_column = self.n_column_spinbox.value()
        exist_column = len(self.x_bias_wl)
        if n_column == exist_column:
            return

        if n_column < exist_column:
            del_index = np.arange(n_column, exist_column)[::-1]
            for da_ind in del_index:
                self.sites_distance_wl[da_ind].sig_text_changed.disconnect(self.sites_distance_changed)
                self.per_max_sites_wl[da_ind].sig_text_changed.disconnect(self.per_max_sites_changed)
                self.x_bias_wl[da_ind].sig_text_changed.disconnect(self.x_bias_changed)
                self.y_bias_wl[da_ind].sig_text_changed.disconnect(self.y_bias_changed)

                self.right_layout.removeWidget(self.sites_distance_wl[da_ind])
                self.right_layout.removeWidget(self.per_max_sites_wl[da_ind])
                self.right_layout.removeWidget(self.x_bias_wl[da_ind])
                self.right_layout.removeWidget(self.y_bias_wl[da_ind])
                self.sites_distance_wl[da_ind].deleteLater()
                self.per_max_sites_wl[da_ind].deleteLater()
                self.x_bias_wl[da_ind].deleteLater()
                self.y_bias_wl[da_ind].deleteLater()
                del self.sites_distance_wl[da_ind]
                del self.per_max_sites_wl[da_ind]
                del self.x_bias_wl[da_ind]
                del self.y_bias_wl[da_ind]

                del self.probe_settings['sites_distance'][da_ind]
                del self.probe_settings['per_max_sites'][da_ind]
                del self.probe_settings['x_bias'][da_ind]
                del self.probe_settings['y_bias'][da_ind]
        else:
            for da_ind in range(exist_column, n_column):
                self.sites_distance_wl.append(IntLineEdit(da_ind, str(0)))
                self.per_max_sites_wl.append(IntLineEdit(da_ind, str(0)))
                self.x_bias_wl.append(IntLineEdit(da_ind, str(0)))
                self.y_bias_wl.append(IntLineEdit(da_ind, str(0)))

                self.sites_distance_wl[-1].sig_text_changed.connect(self.sites_distance_changed)
                self.per_max_sites_wl[-1].sig_text_changed.connect(self.per_max_sites_changed)
                self.x_bias_wl[-1].sig_text_changed.connect(self.x_bias_changed)
                self.y_bias_wl[-1].sig_text_changed.connect(self.y_bias_changed)

                self.right_layout.addWidget(self.sites_distance_wl[-1], 6, da_ind + 1, 1, 1)
                self.right_layout.addWidget(self.per_max_sites_wl[-1], 7, da_ind + 1, 1, 1)
                self.right_layout.addWidget(self.x_bias_wl[-1], 8, da_ind + 1, 1, 1)
                self.right_layout.addWidget(self.y_bias_wl[-1], 9, da_ind + 1, 1, 1)

                self.probe_settings['sites_distance'].append(0)
                self.probe_settings['per_max_sites'].append(0)
                self.probe_settings['x_bias'].append(0)
                self.probe_settings['y_bias'].append(0)

    def probe_length_changed(self, obj):
        text_val = obj[1]
        if text_val in ['', '-', '+']:
            self.button_box.setEnabled(False)
        else:
            self.probe_settings['probe_length'] = int(text_val)
            if float(text_val) < 1e-4:
                self.button_box.setEnabled(False)
            else:
                self.button_box.setEnabled(True)

    def probe_thickness_changed(self, obj):
        text_val = obj[1]
        if text_val in ['', '-', '+']:
            self.button_box.setEnabled(False)
        else:
            self.probe_settings['probe_thickness'] = int(text_val)
            if float(text_val) < 0:
                self.button_box.setEnabled(False)
            else:
                self.button_box.setEnabled(True)

    def tip_length_changed(self, obj):
        text_val = obj[1]
        if text_val in ['', '-', '+']:
            self.button_box.setEnabled(False)
        else:
            self.probe_settings['tip_length'] = int(text_val)
            if float(text_val) < 0 or float(text_val) > self.probe_settings['probe_length']:
                self.button_box.setEnabled(False)
            else:
                self.button_box.setEnabled(True)

    def site_width_changed(self, obj):
        text_val = obj[1]
        if text_val in ['', '-', '+']:
            self.button_box.setEnabled(False)
        else:
            self.probe_settings['site_width'] = int(text_val)
            if float(text_val) < 1e-4:
                self.button_box.setEnabled(False)
            else:
                self.button_box.setEnabled(True)

    def site_height_changed(self, obj):
        text_val = obj[1]
        if text_val in ['', '-', '+']:
            self.button_box.setVisible(False)
        else:
            self.probe_settings['site_height'] = int(text_val)
            if float(text_val) < 1e-4:
                self.button_box.setEnabled(False)
            else:
                self.button_box.setEnabled(True)

    #
    def sites_distance_changed(self, obj):
        w_id = obj[0]
        text_val = obj[1]
        if text_val in ['', '-', '+']:
            self.button_box.setEnabled(False)
        else:
            self.probe_settings['sites_distance'][w_id] = int(text_val)
            if float(text_val) < 1e-4:
                self.button_box.setEnabled(False)
            else:
                self.button_box.setEnabled(True)

    def per_max_sites_changed(self, obj):
        w_id = obj[0]
        text_val = obj[1]
        if text_val in ['', '-', '+']:
            self.button_box.setEnabled(False)
        else:
            self.probe_settings['per_max_sites'][w_id] = int(text_val)
            if float(text_val) < 1e-4:
                self.button_box.setEnabled(False)
            else:
                self.button_box.setEnabled(True)

    def x_bias_changed(self, obj):
        w_id = obj[0]
        text_val = obj[1]
        # text_val = self.x_bias_wl[index].text()
        if text_val in ['', '-', '+']:
            self.button_box.setEnabled(False)
        else:
            self.probe_settings['x_bias'][w_id] = int(text_val)
            self.button_box.setEnabled(True)

    def y_bias_changed(self, obj):
        w_id = obj[0]
        text_val = obj[1]
        if text_val in ['', '-', '+']:
            self.button_box.setEnabled(False)
        else:
            self.probe_settings['y_bias'][w_id] = int(text_val)
            if float(text_val) < 0:
                self.button_box.setEnabled(False)
            else:
                self.button_box.setEnabled(True)


class MultiProbePlanningDialog(QDialog):
    def __init__(self, multi_settings):
        super().__init__()

        self.setWindowTitle('Multi-Probe Geometry Setting Window')
        self.setStyleSheet(dialog_style)

        btn_box = QDialogButtonBox.Ok | QDialogButtonBox.Cancel

        self.button_box = QDialogButtonBox(btn_box)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        fig_label = QLabel()
        pixmap = QPixmap('icons/toolbar/multi_probe.png')
        fig_label.setPixmap(pixmap)
        fig_label.setFixedHeight(300)

        # initial values
        if multi_settings is None:
            self.multi_settings = {}
            self.multi_settings['x_vals'] = [-300, 300, -100, -300, 300]
            self.multi_settings['y_vals'] = [-100, -100, 0, 100, 100]
            self.multi_settings['faces'] = [0, 0, 0, 0, 0]

        else:
            self.multi_settings = multi_settings

        n_probe = len(self.multi_settings['x_vals'])

        # initial widgets
        n_probe_label = QLabel('Number of Probes (stk): ')
        self.n_probe_spinbox = QSpinBox()
        self.n_probe_spinbox.setMinimum(2)
        self.n_probe_spinbox.setValue(n_probe)

        row_names = [QLabel('X Coordinates (um): '), QLabel('Y Coordinates (um): '),
                     QLabel('Probe Faces: ')]

        self.x_val_wl = []
        self.y_val_wl = []
        self.faces_wl = []
        # self.types_wl = []
        for i in range(n_probe):
            self.x_val_wl.append(IntLineEdit(i, str(self.multi_settings['x_vals'][i])))
            self.y_val_wl.append(IntLineEdit(i, str(self.multi_settings['y_vals'][i])))
            self.faces_wl.append(FaceCombo(i))
            self.faces_wl[-1].addItems(['Out', 'In', 'Left', 'Right'])
            self.faces_wl[-1].setCurrentText(str(self.multi_settings['faces'][i]))
            self.x_val_wl[-1].sig_text_changed.connect(self.x_vals_changed)
            self.y_val_wl[-1].sig_text_changed.connect(self.y_vals_changed)
            self.faces_wl[-1].sig_text_changed.connect(self.faces_changed)

        # connect
        self.n_probe_spinbox.valueChanged.connect(self.n_probe_changed)

        # add widgets to layout
        right_frame = QFrame()
        self.right_layout = QGridLayout(right_frame)
        self.right_layout.addWidget(n_probe_label, 0, 0, 1, 1)
        self.right_layout.addWidget(self.n_probe_spinbox, 0, 1, 1, 1)

        for i in range(len(row_names)):
            self.right_layout.addWidget(row_names[i], 1 + i, 0, 1, 1)

        for j in range(n_probe):
            self.right_layout.addWidget(self.x_val_wl[j], 1, j + 1, 1, 1)
            self.right_layout.addWidget(self.y_val_wl[j], 2, j + 1, 1, 1)
            self.right_layout.addWidget(self.faces_wl[j], 3, j + 1, 1, 1)

        #
        out_frame = QFrame()
        out_layout = QHBoxLayout(out_frame)
        out_layout.addWidget(fig_label)
        out_layout.addWidget(right_frame)

        # add widget to layout
        layout = QVBoxLayout()
        layout.addWidget(out_frame)
        layout.addSpacing(10)
        layout.addWidget(self.button_box)
        self.setLayout(layout)

    def n_probe_changed(self):
        n_probe = self.n_probe_spinbox.value()
        exist_probe = len(self.x_val_wl)
        if n_probe == exist_probe:
            return

        if n_probe < exist_probe:
            del_index = np.arange(n_probe, exist_probe)[::-1]
            for da_ind in del_index:
                self.x_val_wl[da_ind].sig_text_changed.disconnect(self.x_vals_changed)
                self.y_val_wl[da_ind].sig_text_changed.disconnect(self.y_vals_changed)
                self.faces_wl[da_ind].sig_text_changed.disconnect(self.faces_changed)

                self.right_layout.removeWidget(self.x_val_wl[da_ind])
                self.right_layout.removeWidget(self.y_val_wl[da_ind])
                self.right_layout.removeWidget(self.faces_wl[da_ind])
                self.x_val_wl[da_ind].deleteLater()
                self.y_val_wl[da_ind].deleteLater()
                self.faces_wl[da_ind].deleteLater()
                del self.x_val_wl[da_ind]
                del self.y_val_wl[da_ind]
                del self.faces_wl[da_ind]

                del self.multi_settings['x_vals'][da_ind]
                del self.multi_settings['y_vals'][da_ind]
                del self.multi_settings['faces'][da_ind]
        else:
            for da_ind in range(exist_probe, n_probe):
                self.x_val_wl.append(IntLineEdit(da_ind, str(0)))
                self.y_val_wl.append(IntLineEdit(da_ind, str(0)))
                self.faces_wl.append(FaceCombo(da_ind))
                self.faces_wl[-1].addItems(['Out', 'In', 'Left', 'Right'])

                self.x_val_wl[-1].sig_text_changed.connect(self.x_vals_changed)
                self.y_val_wl[-1].sig_text_changed.connect(self.y_vals_changed)
                self.faces_wl[-1].sig_text_changed.connect(self.faces_changed)

                self.right_layout.addWidget(self.x_val_wl[-1], 1, da_ind + 1, 1, 1)
                self.right_layout.addWidget(self.y_val_wl[-1], 2, da_ind + 1, 1, 1)
                self.right_layout.addWidget(self.faces_wl[-1], 3, da_ind + 1, 1, 1)

                self.multi_settings['x_vals'].append(0)
                self.multi_settings['y_vals'].append(0)
                self.multi_settings['faces'].append(0)

    #
    def x_vals_changed(self, obj):
        w_id = obj[0]
        text_val = obj[1]
        if text_val in ['', '-', '+']:
            self.button_box.setEnabled(False)
        else:
            self.multi_settings['x_vals'][w_id] = int(text_val)
            self.button_box.setEnabled(True)

    def y_vals_changed(self, obj):
        w_id = obj[0]
        text_val = obj[1]
        if text_val in ['', '-', '+']:
            self.button_box.setEnabled(False)
        else:
            self.multi_settings['y_vals'][w_id] = int(text_val)
            self.button_box.setEnabled(True)

    def faces_changed(self, obj):
        w_id = obj[0]
        index_val = obj[1]
        self.multi_settings['faces'][w_id] = index_val







"""
@author: Jingyi GF{jingyi.g.fuglstad@gmail.com}
"""
import os
import sys
from sys import platform
from setuptools import setup, find_packages


is_problematic = False

if sys.version_info[:2] < (3, 8) or sys.version_info[:2] > (3, 11):
    is_problematic = True

if sys.version_info.minor == 8 and sys.version_info.micro < 10:
    is_problematic = True

if sys.version_info.minor == 9 and sys.version_info.micro == 0:
    is_problematic = True

if sys.version_info.minor == 10 and sys.version_info.micro > 10:
    is_problematic = True

if is_problematic:
    raise RuntimeError("Python version >= 3.8.10 <= 3.10.9 / 3.9.0 required.")


# Utility function to read the README file.
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


CLASSIFIERS = """
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
Intended Audience :: Developers
Intended Audience :: Education
License :: OSI Approved :: MIT License
Programming Language :: Python :: 3
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Programming Language :: Python :: 3.10
Programming Language :: Python :: 3 :: Only
Topic :: Software Development
Topic :: Scientific/Engineering
Typing :: Typed
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""

REQUIRES = """
PyQt5 >= 5.14.2; python_version == "3.8"
PyQt5 >= 5.15.1; python_version == "3.9"
PyQt5 >= 5.15.5; python_version == "3.10"
aicspylibczi >= 3.0.3
pyqtgraph == 0.12.3
PyOpenGL >= 3.1.5
QtRangeSlider == 0.1.5
opencv-python >= 4.5.4.60
numba >= 0.54.1
numpy >= 1.20.3
scipy >= 1.7.3
requests >= 2.26.0
nibabel >= 3.2.1
pynrrd >= 0.4.3
tifffile >= 2021.11.2
pandas >= 1.3.5
natsort >= 8.0.2
imagecodecs >= 2022.2.22
h5py >= 3.7.0
tables >= 3.7.0
"""

PACKAGE_DATA = """
main_window.ui
data/query.csv
data/atlas_labels.pkl
icons/*.svg
icons/*.png
icons/layers/*.svg
icons/layers/*.png
icons/sidebar/*.svg
icons/sidebar/*.png
icons/toolbar/*.svg
icons/toolbar/*.png
qss/*.qss
"""


setup(
    name="herbs",
    version="0.2.8",
    author="Jingyi GF",
    author_email="jingyi.g.fuglstad@gmail.com",
    description="A Python-based GUI for Histological E-data Registration in Brain Space",
    keywords="brain atlas, histological image registration, probe coordinates",
    url="https://github.com/JingyiGF/HERBS",
    packages=find_packages(),
    package_data={"": [_f for _f in PACKAGE_DATA.split("\n") if _f]},
    include_package_data=True,
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    project_urls={
        "Bug Tracker": "https://github.com/JingyiGF/HERBS/issues",
    },
    classifiers=[_f for _f in CLASSIFIERS.split("\n") if _f],
    python_requires=">=3.8",
    install_requires=[_f for _f in REQUIRES.split("\n") if _f],
)



import os
from os.path import dirname, realpath, join
import sys
from sys import argv, exit
from pathlib import Path
import nrrd
import pickle
import csv
import nibabel as nib
import numpy as np
import pandas as pd
import cv2
from PyQt5.QtGui import *
from PyQt5.QtCore import *
from PyQt5.QtWidgets import *

from .uuuuuu import make_contour_img, make_atlas_label_contour
from .obj_items import render_volume, render_small_volume


def _make_label_info_data_waxholm_rat(label_file_path, excel_file_path):
    # label_file_path = '..../WHS_SD_rat_atlas_v4.label'
    # excel_file_path = '..../WHS SD rat brain atlas v4 labels for MBAT.xlsx'

    xl_file = pd.ExcelFile(excel_file_path)
    dfs = {sheet_name: xl_file.parse(sheet_name) for sheet_name in xl_file.sheet_names}
    dfs_keys = list(dfs.keys())
    if len(dfs_keys) != 1:
        raise Exception('need to be only 1 sheet')

    df = dfs[dfs_keys[0]]

    index = []
    level = []
    name = []

    for i in range(df.shape[0]):
        da_line = df.iloc[i].values
        for j in range(df.shape[1]):
            if ~np.isnan(da_line[j]):
                print(da_line[j])
                index.append(da_line[j])
                level.append(j)
                name.append(da_line[j+1])
                break

    index = np.ravel(index).astype(int)
    abv = df['Abbreviation'].values[:len(index)]
    parent = df['Parent'].values[:len(index)].astype(int)

    file = open(label_file_path, 'rb')
    lines = file.readlines()
    file.close()

    for i in range(len(lines)):
        da_line = lines[i].decode()
        if da_line[0] == '#':
            continue
        start_line = i
        break

    lindex = []
    red = []
    green = []
    blue = []
    lname = []
    for i in range(start_line, len(lines)):
        da_line = lines[i].decode()
        print(da_line)
        da_elements = da_line.split('"')
        da_numbers = da_elements[0].split()
        lindex.append(int(da_numbers[0]))
        red.append(int(da_numbers[1]))
        green.append(int(da_numbers[2]))
        blue.append(int(da_numbers[3]))
        lname.append(da_elements[1])

    lindex = np.ravel(lindex)
    red = np.ravel(red)
    green = np.ravel(green)
    blue = np.ravel(blue)

    colors = np.zeros((len(index), 3))
    colors[:] = np.nan

    for i in range(1, len(lname)):
        print(i)
        if lindex[i] not in index:
            raise Exception('not matching')

    for i in range(len(index)):
        if index[i] > 600:
            if index[i] == 1000:
                colors[i] = np.array([50, 168, 82])
            elif index[i] in [1001, 1050, 1002, 1003, 1004, 1005, 1006, 1051, 1007, 1008, 1009, 1010, 1011, 1012]:
                colors[i] = np.array([255, 255, 255])
            elif index[i] == 1048:
                colors[i] = np.array([114, 126, 186])
            elif index[i] == 1049:
                colors[i] = np.array([16, 79, 24])
            else:
                colors[i] = np.array([128, 128, 128])
        else:
            da_ind = np.where(lindex == index[i])[0][0]
            colors[i] = np.array([red[da_ind], green[da_ind], blue[da_ind]])

    label = {}
    label['index'] = index
    label['color'] = colors.astype(int)
    label['label'] = name
    label['abbrev'] = abv
    label['parent'] = parent
    label['level_indicator'] = np.ravel(level)

    outfile = open('atlas_labels.pkl', 'wb')
    pickle.dump(label, outfile)
    outfile.close()


def check_data_path_and_load(file_path):
    success = True
    filename, file_extension = os.path.splitext(file_path)
    if file_extension == '.pkl':
        try:
            infile = open(file_path, 'rb')
            data = pickle.load(infile)
            infile.close()
        except (IndexError, AttributeError, IOError, OSError):
            success = False
    elif file_extension == '.nrrd':
        try:
            data, header = nrrd.read(file_path)
        except (IndexError, AttributeError, IOError, OSError):
            success = False
    else:
        try:
            data_file = nib.load(file_path)
            data = data_file.get_fdata()
        except (IndexError, AttributeError, IOError, OSError):
            success = False
    return data, success



def check_atlas_file_path(atlas_folder, data_file=None, segmentation_file=None):
    atlas_path = os.path.join(atlas_folder, data_file)
    segmentation_path = os.path.join(atlas_folder, segmentation_file)

    if not os.path.exists(atlas_path) or not os.path.exists(segmentation_path):
        msg = 'atlas_path or segmentation_path not exist.'
        msg_flag = 0
    else:
        msg = 'atlas path and segmentation path exist.'
        msg_flag = 1

    return msg, msg_flag, atlas_path, segmentation_path


def load_mask_file(atlas_folder, mask_file=None):
    if mask_file is not None:
        mask_path = os.path.join(atlas_folder, mask_file)
        if os.path.exists(mask_path):
            mask_data, mask_success = check_data_path_and_load(mask_path)
            if not mask_success:
                msg = 'Failed to load mask data.'
                msg_flag = 0
                mask_data = None
            else:
                msg = 'Mask data loaded successfully.'
                msg_flag = 1
                mask_data = mask_data[:, :, :, 0]  # ????? check data again when you have time, dear
        else:
            msg = 'Mask path not exist.'
            msg_flag = 0
            mask_data = None
    else:
        msg = 'No mask data is needed.'
        msg_flag = 1
        mask_data = None

    return msg, msg_flag, mask_data


def process_segmentation_data(atlas_folder, segmentation_path, mask_data):
    segmentation_data, seg_success = check_data_path_and_load(segmentation_path)
    if not seg_success:
        msg = 'Failed to load segmentation data.'
        msg_flag = 0
    else:
        if mask_data is not None:
            # make segmentation with mask
            for i in range(len(mask_data)):
                segmentation_data[i][mask_data[i] == 0] = 0
            segmentation_data = segmentation_data.astype('int')

        unique_label = np.unique(segmentation_data)

        segment = {'data': segmentation_data, 'unique_label': unique_label}

        outfile = open(os.path.join(atlas_folder, 'segment_pre_made.pkl'), 'wb')
        pickle.dump(segment, outfile)
        outfile.close()

        msg = 'Segmentation data processed successfully.'
        msg_flag = 1

    return msg, msg_flag


def process_atlas_data(atlas_folder, atlas_path, mask_data,
                       bregma_coordinates=None, lambda_coordinates=None, voxel_size=None):
    atlas_data, atlas_success = check_data_path_and_load(atlas_path)
    if not atlas_success:
        msg = 'Failed to load atlas volume data.'
        msg_flag = 0
    else:

        new_atlas_data = atlas_data.copy()
        if mask_data is not None:
            # make atlas with mask
            new_atlas_data = new_atlas_data - np.min(new_atlas_data)
            for i in range(len(mask_data)):
                new_atlas_data[i][mask_data[i] == 0] = 0
            new_atlas_data = new_atlas_data / np.max(new_atlas_data)
        else:
            new_atlas_data = new_atlas_data - np.min(new_atlas_data)
            new_atlas_data = new_atlas_data / np.max(new_atlas_data)

        atlas_info = [
            {'name': 'anterior', 'values': np.arange(new_atlas_data.shape[0]) * voxel_size, 'units': 'um'},
            {'name': 'dorsal', 'values': np.arange(new_atlas_data.shape[1]) * voxel_size, 'units': 'um'},
            {'name': 'right', 'values': np.arange(new_atlas_data.shape[2]) * voxel_size, 'units': 'um'},
            {'vxsize': voxel_size,
             'Bregma': [bregma_coordinates[0], bregma_coordinates[1], bregma_coordinates[2]],
             'Lambda': [lambda_coordinates[0], lambda_coordinates[1], lambda_coordinates[2]]}
        ]

        atlas = {'data': new_atlas_data, 'info': atlas_info}

        atlas_data = atlas['data']
        atlas_info = atlas['info']

        outfile = open(os.path.join(atlas_folder, 'atlas_pre_made.pkl'), 'wb')
        pickle.dump(atlas, outfile)
        outfile.close()

        msg = 'Volume Atlas data processed successfully.'
        msg_flag = 1

    return msg, msg_flag


def process_contour_data(segmentation_data, dim_index=0):
    contour_img = np.zeros(segmentation_data.shape, 'i')

    # pre-process boundary
    if dim_index == 0:
        for i in range(segmentation_data.shape[dim_index]):
            da_slice = segmentation_data[i, :, :].copy()
            da_contour = make_contour_img(da_slice)
            contour_img[i, :, :] = da_contour
    elif dim_index == 1:
        for i in range(segmentation_data.shape[dim_index]):
            da_slice = segmentation_data[:, i, :].copy()
            da_contour = make_contour_img(da_slice)
            contour_img[:, i, :] = da_contour
    else:
        for i in range(segmentation_data.shape[dim_index]):
            da_slice = segmentation_data[:, :, i].copy()
            da_contour = make_contour_img(da_slice)
            contour_img[:, :, i] = da_contour
    return contour_img



# boundary = {'s_contour': sagital_contour_img,
#                 'c_contour': coronal_contour_img,
#                 'h_contour': horizontal_contour_img}
#
#     bnd = {'data': boundary}
#
#     outfile_ct = open(os.path.join(atlas_folder, 'contour_pre_made.pkl'), 'wb')
#     pickle.dump(bnd, outfile_ct)
#     outfile_ct.close()

def process_atlas_raw_data(atlas_folder, data_file=None, segmentation_file=None, mask_file=None,
                           bregma_coordinates=None, lambda_coordinates=None, voxel_size=None):
    atlas_data, atlas_info, segmentation_data, boundary = None, None, None, None

    atlas_path = os.path.join(atlas_folder, data_file)
    segmentation_path = os.path.join(atlas_folder, segmentation_file)

    if mask_file is not None:
        mask_path = os.path.join(atlas_folder, mask_file)
        if os.path.exists(mask_path):
            mask_data, mask_success = check_data_path_and_load(mask_path)
            if not mask_success:
                msg = 'Failed to load mask data.'
                return msg
            mask_data = mask_data[:, :, :, 0]  # ????? check data again when you have time, dear
        else:
            mask_data = None
    else:
        mask_data = None

    if not os.path.exists(atlas_path) or not os.path.exists(segmentation_path):
        msg = 'atlas_path or segmentation_path not exist.'
        return msg

    # pre-process segmentation
    segmentation_data, seg_success = check_data_path_and_load(segmentation_path)
    if not seg_success:
        msg = 'Failed to load segmentation data.'
        return msg

    if mask_data is not None:
        # make segmentation with mask
        for i in range(len(mask_data)):
            segmentation_data[i][mask_data[i] == 0] = 0
        segmentation_data = segmentation_data.astype('int')

    unique_label = np.unique(segmentation_data)

    segment = {'data': segmentation_data, 'unique_label': unique_label}

    outfile = open(os.path.join(atlas_folder, 'segment_pre_made.pkl'), 'wb')
    pickle.dump(segment, outfile)
    outfile.close()

    # pre-process atlas
    atlas_data, atlas_success = check_data_path_and_load(atlas_path)
    if not atlas_success:
        msg = 'Failed to load atlas volume data.'
        return msg

    new_atlas_data = atlas_data.copy()
    if mask_data is not None:
        # make atlas with mask
        new_atlas_data = new_atlas_data - np.min(new_atlas_data)
        for i in range(len(mask_data)):
            new_atlas_data[i][mask_data[i] == 0] = 0
        new_atlas_data = new_atlas_data / np.max(new_atlas_data)
    else:
        new_atlas_data = new_atlas_data - np.min(new_atlas_data)
        new_atlas_data = new_atlas_data / np.max(new_atlas_data)

    atlas_info = [
        {'name': 'anterior', 'values': np.arange(new_atlas_data.shape[0]) * voxel_size, 'units': 'um'},
        {'name': 'dorsal', 'values': np.arange(new_atlas_data.shape[1]) * voxel_size, 'units': 'um'},
        {'name': 'right', 'values': np.arange(new_atlas_data.shape[2]) * voxel_size, 'units': 'um'},
        {'vxsize': voxel_size,
         'Bregma': [bregma_coordinates[0], bregma_coordinates[1], bregma_coordinates[2]],
         'Lambda': [lambda_coordinates[0], lambda_coordinates[1], lambda_coordinates[2]]}
    ]

    atlas = {'data': new_atlas_data, 'info': atlas_info}

    atlas_data = atlas['data']
    atlas_info = atlas['info']

    outfile = open(os.path.join(atlas_folder, 'atlas_pre_made.pkl'), 'wb')
    pickle.dump(atlas, outfile)
    outfile.close()

    boundary = make_atlas_label_contour(atlas_folder, segmentation_data)

    msg = 'Atlas loaded successfully.'

    return atlas_data, atlas_info, segmentation_data, unique_label, boundary, msg






class AtlasMeshProcessor(object):
    def __init__(self, atlas_folder, atlas_data, segmentation_data, factor, level):
        meshdata = render_volume(atlas_data, atlas_folder, factor=factor, level=level)

        small_meshdata_list = render_small_volume(atlas_data, segmentation_data, atlas_folder,
                                                  factor=factor, level=level)


class AtlasLoader(object):
    def __init__(self, atlas_folder):
        pre_made_atlas_path = os.path.join(atlas_folder, 'atlas_pre_made.pkl')
        pre_made_segment_path = os.path.join(atlas_folder, 'segment_pre_made.pkl')
        pre_made_boundary_path = os.path.join(atlas_folder, 'contour_pre_made.pkl')

        pre_s_boundary_path = os.path.join(atlas_folder, 'sagital_contour_pre_made.pkl')
        pre_c_boundary_path = os.path.join(atlas_folder, 'coronal_contour_pre_made.pkl')
        pre_h_boundary_path = os.path.join(atlas_folder, 'horizontal_contour_pre_made.pkl')

        pre_made_label_info_path = os.path.join(atlas_folder, 'atlas_labels.pkl')

        chk1 = os.path.exists(pre_made_label_info_path)
        chk2 = os.path.exists(pre_made_atlas_path)
        chk3 = os.path.exists(pre_made_segment_path)

        if not np.all([chk1, chk2, chk3]):
            self.msg = 'Please pre-process the raw data of your desire atlas.'
            self.success = False
        else:
            # laod label info
            try:
                infile = open(pre_made_label_info_path, 'rb')
                self.label_info = pickle.load(infile)
                infile.close()
                self.success = True
            except ValueError:
                self.msg = 'Please give the label information file. If you do not have one, ' \
                           'the maintainers are more than happy to help you to process one.'
                self.success = False
            # load atlas
            try:
                infile_atlas = open(pre_made_atlas_path, 'rb')
                atlas = pickle.load(infile_atlas)
                infile_atlas.close()

                self.atlas_data = atlas['data']
                self.atlas_info = atlas['info']
                # load segmentation
                infile_seg = open(pre_made_segment_path, 'rb')
                segment = pickle.load(infile_seg)
                infile_seg.close()

                self.segmentation_data = segment['data']

                data_shape = self.segmentation_data.shape
                print(data_shape)

                self.unique_label = segment['unique_label']
                self.success = True
            except (ValueError, pickle.UnpicklingError):
                self.msg = 'Please re-process atlas and label segmentation file.'
                self.success = False

        bnd_chk0 = os.path.exists(pre_made_boundary_path)
        bnd_chk1 = os.path.exists(pre_s_boundary_path)
        bnd_chk2 = os.path.exists(pre_c_boundary_path)
        bnd_chk3 = os.path.exists(pre_h_boundary_path)

        if not bnd_chk0:
            if not np.all([bnd_chk1, bnd_chk2, bnd_chk3]):
                self.msg = 'Please pre-process boundary file.'
                self.success = False
            else:
                # load boundary
                self.boundary = {}
                try:
                    infile = open(pre_s_boundary_path, 'rb')
                    sagital_contour_img = pickle.load(infile)
                    infile.close()

                    self.boundary['s_contour'] = sagital_contour_img
                    self.success = True
                except (ValueError, pickle.UnpicklingError, BrokenPipeError):
                    self.msg = 'Please re-process boundary file.'
                    self.success = False

                try:
                    infile = open(pre_c_boundary_path, 'rb')
                    coronal_contour_img = pickle.load(infile)
                    infile.close()

                    self.boundary['c_contour'] = coronal_contour_img
                    self.success = True
                except (ValueError, pickle.UnpicklingError, BrokenPipeError):
                    self.msg = 'Please re-process boundary file.'
                    self.success = False

                try:
                    infile = open(pre_h_boundary_path, 'rb')
                    horizontal_contour_img = pickle.load(infile)
                    infile.close()

                    self.boundary['h_contour'] = horizontal_contour_img
                    self.success = True
                except (ValueError, pickle.UnpicklingError, BrokenPipeError):
                    self.msg = 'Please re-process boundary file.'
                    self.success = False
        else:
            try:
                infile = open(pre_made_boundary_path, 'rb')
                bnd = pickle.load(infile)
                infile.close()

                self.boundary = bnd['data']
                self.success = True
            except (ValueError, pickle.UnpicklingError, BrokenPipeError):
                self.msg = 'Please re-process boundary file.'
                self.success = False











class AtlasMeshLoader(object):
    def __init__(self, atlas_folder):
        pre_made_meshdata_path = os.path.join(atlas_folder, 'atlas_meshdata.pkl')




import os
import pickle
import pyqtgraph as pg
import pyqtgraph.opengl as gl
import numpy as np
import scipy.ndimage as ndi


def get_object_vis_color(color):
    vis_color = (color[0] / 255, color[1] / 255, color[2] / 255, 1)
    return vis_color


def create_plot_points_in_3d(data_dict):
    data_list = data_dict['data']
    pnts = data_list[0]
    for i in range(1, len(data_list)):
        pnts = np.vstack([pnts, data_list[i]])
    vis_color = get_object_vis_color(data_dict['vis_color'])
    vis_points = gl.GLScatterPlotItem(pos=pnts, color=vis_color, size=3)
    return vis_points


def create_probe_line_in_3d(data_dict):
    pos = np.stack([data_dict['insertion_coords_3d'], data_dict['terminus_coords_3d']], axis=0)
    vis_color = get_object_vis_color(data_dict['vis_color'])
    probe_line = gl.GLLinePlotItem(pos=pos, color=vis_color, width=2, mode='line_strip')
    return probe_line


def create_drawing_in_3d(data_dict):
    data_list = data_dict['data']
    pos = data_list[0]
    for i in range(1, len(data_list)):
        pos = np.vstack([pos, data_list[i]])

    vis_color = get_object_vis_color(data_dict['vis_color'])
    if data_dict['plot_mode'] == 'area':
        drawing_obj = gl.GLScatterPlotItem(pos=pos, color=vis_color, size=3, pxMode=True)
    else:
        drawing_obj = gl.GLLinePlotItem(pos=pos, color=vis_color, width=2, mode='line_strip')
    return drawing_obj


def create_contour_line_in_3d(data_dict):
    data_list = data_dict['data']
    pos = data_list[0]
    for i in range(1, len(data_list)):
        pos = np.vstack([pos, data_list[i]])
    if np.any(pos[0] != pos[-1]):
        pos = np.vstack([pos, pos[0]])

    vis_color = get_object_vis_color(data_dict['vis_color'])
    contour_obj = gl.GLLinePlotItem(pos=pos, color=vis_color, width=2, mode='line_strip')
    return contour_obj


def make_3d_gl_widget(data_dict, obj_type):
    if obj_type == 'merged probe':
        obj_3d = create_probe_line_in_3d(data_dict)
    elif obj_type == 'merged virus':
        obj_3d = create_plot_points_in_3d(data_dict)
    elif obj_type == 'merged cells':
        obj_3d = create_plot_points_in_3d(data_dict)
    elif obj_type == 'merged contour':
        obj_3d = create_contour_line_in_3d(data_dict)
    else:
        obj_3d = create_drawing_in_3d(data_dict)
    return obj_3d


def render_volume(atlas_data, atlas_folder, factor=2, level=0.1):
    da_data = atlas_data.copy()
    img = np.ascontiguousarray(da_data[::factor, ::factor, ::factor])
    verts, faces = pg.isosurface(ndi.gaussian_filter(img.astype('float64'), (2, 2, 2)), np.max(da_data) * level)

    md = gl.MeshData(vertexes=verts * factor, faces=faces)

    outfile = open(os.path.join(atlas_folder, 'atlas_meshdata.pkl'), 'wb')
    pickle.dump(md, outfile)
    outfile.close()

    return md


def render_small_volume(label_id, save_path, atlas_data, atlas_label, factor=2, level=0.1):
    temp_atlas = atlas_data.copy()
    temp_atlas[atlas_label != label_id] = 0
    pimg = np.ascontiguousarray(temp_atlas[::factor, ::factor, ::factor])
    verts, faces = pg.isosurface(ndi.gaussian_filter(pimg.astype('float64'), (2, 2, 2)), np.max(temp_atlas) * level)
    # small_verts_list[str(id)] = verts
    # small_faces_list[str(id)] = faces

    md = gl.MeshData(vertexes=verts * factor, faces=faces)

    outfile = open(os.path.join(save_path, '{}.pkl'.format(label_id)), 'wb')
    pickle.dump(md, outfile)
    outfile.close()





import os
import sys
import numpy as np
import pyqtgraph as pg
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
from PyQt5.QtCore import *
from pyqtgraph.Qt import QtGui, QtCore

from qtrangeslider import QRangeSlider
import scipy.interpolate
# from scipy.interpolate import interp1d

import cv2

from .image_stacks import ImageStacks
from .image_reader import ImageReader
from .uuuuuu import hsv2rgb, gamma_line, get_qhsv_from_czi_hsv, make_hist_data, read_qss_file
from .widgets_utils import BWSpin, GammaSpin
from .movable_points import MovablePoints


reset_button_style = '''
QPushButton{
    background: #656565;
    border-radius: 5px;
    color: white;
    border-style: outset;
    border-bottom: 1px solid rgb(30, 30, 30);
    height: 24px;
    margin: 0px;
}

QPushButton:hover{
    background-color: #323232;
    border: 1px solid #656565;
}

'''


class CurvesPlot(pg.PlotWidget):
    class SignalProxy(QObject):
        sigLineChange = pyqtSignal(object)
        sigBoundChange = pyqtSignal()

    def __init__(self, parent=None, background='w'):
        self._sigprox = CurvesPlot.SignalProxy()
        self.sig_line_change = self._sigprox.sigLineChange
        self.sig_spin_max_change = self._sigprox.sigBoundChange

        pg.PlotWidget.__init__(self)
        self.setBackground('w')
        self.plotItem.vb.setMouseEnabled(x=False, y=False)
        self.getPlotItem().hideAxis('bottom')
        self.getPlotItem().hideAxis('left')

        self.scene().sigMouseMoved.connect(self.on_mouse_moved)
        self.scene().sigMouseClicked.connect(self.on_mouse_clicked)

        self.adding_allowed = False
        self.enable_channel = [False, False, False, False]
        self.line_type = 'gamma'
        self.depth_level = 65535
        self.start_point = [np.array([0, 0]) for i in range(4)]
        self.end_point = [np.array([self.depth_level, self.depth_level]) for i in range(4)]
        self.table_input = np.arange(self.depth_level + 1)
        self.hist_data = None
        self.active_index = None

        self.inactive_pen = pg.mkPen(QColor(128, 128, 128, 255), width=3)
        self.inactive_brush = QColor(128, 128, 128, 10)

        self.active_pen = []
        self.active_brush = []

        self.hist_list = []
        self.lut_line = []
        self.lut_points = []
        for i in range(4):
            self.hist_list.append(pg.PlotDataItem())
            self.lut_line.append(pg.PlotCurveItem(fillLevel=None))
            self.lut_points.append(MovablePoints())
            self.lut_points[i].mouseDragged.connect(self.on_mouse_dragged)

        for i in range(4):
            self.addItem(self.hist_list[i])
        for i in range(4):
            self.addItem(self.lut_line[i])
            self.lut_line[i].setVisible(False)
        for i in range(4):
            self.addItem(self.lut_points[i])
            self.lut_points[i].setVisible(False)

    def set_data(self, hist_data, color, depth_level):
        """
        Set data to curve Plot view, only called once there load new data
        :param hist_data: grayscale image data for one scene
        :param color: list of rgb [0, 255] for each channel
        :param depth_level: gray intensity level
        :return: None
        """
        if self.hist_data is not None:
            for i in range(4):
                self.hist_list[i].clear()
                self.lut_line[i].clear()
                self.lut_points[i]. clear()
            self.hist_data = None

        if self.depth_level != depth_level:
            self.start_point = [np.array([0, 0]) for i in range(4)]
            self.end_point = [np.array([depth_level, depth_level]) for i in range(4)]
            self.table_input = np.arange(0, depth_level + 1)
            self.depth_level = depth_level

        self.hist_data = hist_data.copy()

        self.plotItem.vb.setRange(xRange=[0, depth_level], yRange=[0, depth_level])

        for i in range(len(self.hist_data)):
            self.enable_channel[i] = True
            self.active_pen.append(pg.mkPen(QColor(color[i][0], color[i][1], color[i][2], 255), width=3))
            self.active_brush.append((color[i][0], color[i][1], color[i][2], 130))
            self.hist_list[i].setVisible(True)
            self.hist_list[i].setData(self.hist_data[i][0], self.hist_data[i][1],
                                      fillLevel=0, fillBrush=self.active_brush[i], pen=self.active_pen[i])
            self.start_point[i][0] = np.min(self.hist_data[i][0])
            self.end_point[i][0] = np.max(self.hist_data[i][0])

            self.set_original_lut_line(i)

            self.lut_points[i].setData(pos=np.vstack([self.start_point[i], self.end_point[i]]))

    def set_original_lut_line(self, ind):
        line_x = np.array([self.start_point[ind][0], self.end_point[ind][0]])
        line_y = np.array([self.start_point[ind][1], self.end_point[ind][1]])
        self.lut_line[ind].setData(line_x, line_y)
        self.lut_line[ind].setPen(self.active_pen[ind])
        self.lut_line[ind].setVisible(True)

    def reset_plot(self):
        for i in range(len(self.hist_data)):
            self.set_original_lut_line(i)
            self.lut_points[i].setData(pos=np.vstack([self.start_point[i], self.end_point[i]]))

    def set_enable(self, ind, is_enable):
        self.enable_channel[ind] = is_enable
        if is_enable:
            self.hist_list[ind].setPen(self.active_pen[ind])
            self.hist_list[ind].setFillBrush(self.active_brush[ind])
            self.lut_line[ind].setPen(self.active_pen[ind])
        else:
            self.hist_list[ind].setPen(self.inactive_pen)
            self.hist_list[ind].setFillBrush(self.inactive_brush)
            self.lut_line[ind].setPen(self.inactive_pen)
        if np.sum(self.enable_channel) == 1 and self.line_type != 'gamma':
            self.active_index = np.where(self.enable_channel)[0][0]
            self.lut_points[self.active_index].setVisible(True)
        else:
            if self.active_index is not None:
                self.lut_points[self.active_index].setVisible(False)
            self.active_index = None

    def change_hist_color(self, color: tuple, ind: int):
        """
        :param color: (h [0, 1], s [0, 1], v [0, 1])
        :param ind: channel index
        :return: None
        """
        if not self.hist_data:
            return
        da_hsv = get_qhsv_from_czi_hsv(color)
        col = QColor.fromHsv(da_hsv[0], da_hsv[1], da_hsv[2], 255)
        self.active_pen[ind] = pg.mkPen(col, width=3)
        self.active_brush[ind] = QColor.fromHsv(da_hsv[0], da_hsv[1], da_hsv[2], 130)
        self.hist_list[ind].setFillBrush(self.active_brush[ind])
        self.hist_list[ind].setPen(self.active_pen[ind])
        self.lut_line[ind].setPen(self.active_pen[ind])

    def change_line_type(self, line_type):
        self.line_type = line_type
        self.reset_plot()

    def set_lut_line(self, table_output, ind):
        self.lut_line[ind].setData(self.table_input, table_output)

    def set_lut_points(self, point_data, ind):
        self.lut_points[ind].setData(pos=point_data)

    def set_plot(self, points, table):
        # self.pnts = points
        # self.table_output = table
        self.line.setData(self.table_input, table)
        self.points.setData(pos=points)

    def on_mouse_dragged(self, vec):
        if self.line_type == 'gamma' or self.active_index is None:
            return
        ind = vec[1]
        data = self.lut_points[self.active_index].data['pos']
        if ind in [0, len(data)-1]:
            return
        else:
            table_output = self.update_table(data)
            self.lut_line[self.active_index].setData(self.table_input, table_output)
        self.sig_line_change.emit((table_output.astype('int'), self.active_index))
        # self.pnts = data.copy()

    def on_mouse_clicked(self, event):
        if self.line_type == 'gamma' or self.active_index is None:
            return
        pnt = self.plotItem.vb.mapSceneToView(event.scenePos())
        data = self.lut_points[self.active_index].data['pos']
        if self.adding_allowed:
            x = pnt.x()
            y = pnt.y()
            temp = np.vstack([data, np.array([x, y])])
            sort_x = np.argsort(temp[:, 0])
            temp = temp[sort_x, :]
            self.lut_points[self.active_index].setData(pos=temp)
            table_output = self.update_table(temp)
            self.lut_line[self.active_index].setData(self.table_input, table_output)
            self.sig_line_change.emit((table_output.astype('int'), self.active_index))

    def on_mouse_moved(self, point):
        if self.line_type == 'gamma' or self.active_index is None:
            return
        pnts = self.plotItem.vb.mapSceneToView(point)
        data = self.lut_points[self.active_index].data['pos']
        sct = self.lut_points[self.active_index].scatter.pointsAt(pnts)
        if self.lut_line[self.active_index].mouseShape().contains(pnts):
            if len(sct) == 0:
                if data[0, 0] < pnts.x() < data[-1, 0]:
                    self.setCursor(Qt.CrossCursor)
                    self.adding_allowed = True
                else:
                    self.setCursor(Qt.ArrowCursor)
                    self.adding_allowed = False
            else:
                self.setCursor(Qt.OpenHandCursor)
                self.adding_allowed = False
        else:
            self.adding_allowed = False
            if len(sct) == 0:
                self.setCursor(Qt.ArrowCursor)
            else:
                self.setCursor(Qt.OpenHandCursor)

    def update_table(self, data):
        cind = np.logical_and(self.table_input >= data[0, 0], self.table_input <= data[-1, 0])
        temp_output = self.table_input.copy()
        if self.line_type == 'linear':
            line_func = scipy.interpolate.interp1d(data[:, 0], data[:, 1], kind='linear')
            temp_output[cind] = line_func(self.table_input[cind])
        else:
            if len(data) < 4:
                line_func = scipy.interpolate.interp1d(data[:, 0], data[:, 1], kind='linear')
                temp_output[cind] = line_func(self.table_input[cind])
            else:
                tck = scipy.interpolate.splrep(data[:, 0], data[:, 1], s=0)
                temp_output[cind] = scipy.interpolate.splev(self.table_input[cind], tck, der=0)

        temp_output[self.table_input <= data[0, 0]] = 0
        temp_output[self.table_input >= data[-1, 0]] = self.depth_level
        temp_output[temp_output <= 0] = 0
        temp_output[temp_output >= self.depth_level] = self.depth_level
        return temp_output


class CurveWidget(QWidget):
    class SignalProxy(QObject):
        sigTableChanged = pyqtSignal(object)
        sigReSet = pyqtSignal()
        sigLineTypeChanged = pyqtSignal()

    def __init__(self, parent=None):
        self._sigprox = CurveWidget.SignalProxy()
        self.sig_table_changed = self._sigprox.sigTableChanged
        self.sig_reset = self._sigprox.sigReSet
        self.sig_line_type_changed = self._sigprox.sigLineTypeChanged

        QWidget.__init__(self)

        self.gray_max = 65535
        self.gamma = [1, 1, 1, 1]
        self.n_channels = None
        self.table_output = []
        self.original_table = []

        self.current_black_val = 0
        self.current_white_val = self.gray_max
        self.current_width = self.gray_max
        self.vis_hist = True

        self.curve_plot = CurvesPlot()
        self.curve_plot.sig_line_change.connect(self.table_changed)
        # self.curve_plot.sigBoundChange.connect(self.spinbox_bound_changed)
        multi_handle_slider_style = read_qss_file('qss/multi_handle_slider.qss')
        self.multi_handle_slider = QRangeSlider(Qt.Horizontal)
        self.multi_handle_slider.setStyleSheet(multi_handle_slider_style)
        self.multi_handle_slider.setMinimum(0)
        self.multi_handle_slider.setMaximum(65535)
        self.multi_handle_slider.setValue((0, 65535))
        self.multi_handle_slider.setSingleStep(1.)
        self.multi_handle_slider.setBarVisible(False)
        self.multi_handle_slider.sliderMoved.connect(self.slider_changed)

        self.black_spinbox = BWSpin()
        self.black_spinbox.setFixedWidth(60)
        self.black_spinbox.spin_nam.setText('Black:')
        self.black_spinbox.spin_val.setMinimum(0)
        self.black_spinbox.spin_val.setMaximum(65535)
        self.black_spinbox.spin_val.setValue(0)
        self.black_spinbox.spin_val.valueChanged.connect(self.black_spinbox_changed)

        self.white_spinbox = BWSpin()
        self.white_spinbox.setFixedWidth(60)
        self.white_spinbox.spin_nam.setText('White:')
        self.white_spinbox.spin_val.setMinimum(0)
        self.white_spinbox.spin_val.setMaximum(65535)
        self.white_spinbox.spin_val.setValue(65535)
        self.white_spinbox.spin_val.valueChanged.connect(self.white_spinbox_changed)

        self.gamma_spinbox = GammaSpin()
        self.gamma_spinbox.setFixedWidth(60)
        self.gamma_spinbox.spin_nam.setText('Gamma:')
        self.gamma_spinbox.spin_val.setMinimum(0.01)
        self.gamma_spinbox.spin_val.setMaximum(16.00)
        self.gamma_spinbox.spin_val.setSingleStep(0.05)
        self.gamma_spinbox.spin_val.setValue(1.00)
        self.gamma_spinbox.spin_val.valueChanged.connect(self.gamma_spinbox_changed)

        self.reset_btn = QPushButton('Reset')
        self.reset_btn.setStyleSheet(reset_button_style)
        self.reset_btn.setFixedHeight(25)
        self.reset_btn.clicked.connect(self.reset_pressed)

        self.line_type_combo = QComboBox()
        self.line_type_combo.setFixedHeight(22)
        ltypes = ['gamma', 'linear', 'spline']
        self.line_type_combo.addItems(ltypes)
        self.line_type_combo.setCurrentText('gamma')
        self.line_type = 'gamma'
        self.line_type_combo.currentTextChanged.connect(self.line_type_changed)

        top_wrap = QFrame()
        top_layout = QGridLayout(top_wrap)
        top_layout.setContentsMargins(0, 0, 0, 0)
        top_layout.addWidget(self.line_type_combo, 0, 0, 1, 1)
        top_layout.addWidget(self.reset_btn, 0, 1, 1, 1)

        bgw_wrap = QFrame()
        bgw_wrap_layout = QHBoxLayout(bgw_wrap)
        bgw_wrap_layout.setContentsMargins(0, 0, 0, 0)
        bgw_wrap_layout.addWidget(self.black_spinbox, alignment=Qt.AlignLeft)
        bgw_wrap_layout.addWidget(self.gamma_spinbox, alignment=Qt.AlignCenter)
        bgw_wrap_layout.addWidget(self.white_spinbox, alignment=Qt.AlignRight)

        slider_wrap = QFrame()
        slider_wrap_layout = QVBoxLayout(slider_wrap)
        slider_wrap_layout.setContentsMargins(0, 0, 0, 0)
        slider_wrap_layout.setSpacing(0)
        slider_wrap_layout.addWidget(self.multi_handle_slider)

        bottom_wrap = QFrame()
        bottom_layout = QVBoxLayout(bottom_wrap)
        bottom_layout.setContentsMargins(0, 0, 0, 0)
        bottom_layout.addWidget(slider_wrap)
        bottom_layout.addWidget(bgw_wrap)

        widget_layout = QVBoxLayout()
        widget_layout.setContentsMargins(0, 0, 0, 0)
        widget_layout.addWidget(top_wrap)
        widget_layout.addWidget(self.curve_plot)
        widget_layout.addWidget(bottom_wrap)
        self.setLayout(widget_layout)

    def block_signal(self, is_clocked):
        self.multi_handle_slider.blockSignals(is_clocked)
        self.white_spinbox.spin_val.blockSignals(is_clocked)
        self.black_spinbox.spin_val.blockSignals(is_clocked)

    def set_data(self, data, color, depth_level):

        if self.table_output:
            self.table_output = []

        if self.gray_max != depth_level:
            self.gray_max = depth_level
            self.multi_handle_slider.setMaximum(depth_level)
            self.white_spinbox.spin_val.setMaximum(depth_level)

        self.n_channels = data.shape[2]

        hist_data = make_hist_data(data, depth_level)
        self.curve_plot.set_data(hist_data, color, depth_level)

        for i in range(self.n_channels):
            point_data = np.asarray([self.curve_plot.start_point[i], self.curve_plot.end_point[i]])
            table = self.get_table(point_data, i)
            self.table_output.append(table.astype(int))
        self.original_table = self.table_output.copy()

        self.set_original_to_slider()

    def set_original_to_slider(self):
        min_slider_val = np.min(np.asarray(self.curve_plot.start_point)[:self.n_channels, 0])
        max_slider_val = np.max(np.asarray(self.curve_plot.end_point)[:self.n_channels, 0])
        self.block_signal(True)
        self.multi_handle_slider.setValue((min_slider_val, max_slider_val))
        self.black_spinbox.spin_val.setValue(min_slider_val)
        self.white_spinbox.spin_val.setValue(max_slider_val)
        self.block_signal(False)
        self.current_black_val = min_slider_val
        self.current_white_val = max_slider_val
        self.current_width = max_slider_val - min_slider_val

    def slider_changed(self, ev):
        if self.curve_plot.hist_data is None:
            return
        # point_values = self.curve_plot.lut_points.data['pos'].copy()
        # point_values = point_values.astype(int)
        # self.black_spinbox.spin_val.setMaximum(point_values[1, 0] - 1)
        # self.white_spinbox.spin_val.setMinimum(point_values[-2, 0] + 1)
        val1 = int(ev[0])
        val2 = int(ev[1])

        if val1 != self.black_spinbox.spin_val.value():
            # if val1 >= point_values[1, 0]:
            #     val1 = point_values[1, 0] - 1
            #     self.multi_handle_slider.setValue((val1, val2))
            self.black_spinbox.spin_val.setValue(val1)
        if val2 != self.white_spinbox.spin_val.value():
            # if val2 <= point_values[-2, 0]:
            #     val2 = point_values[-2, 0] + 1
            #     self.multi_handle_slider.setValue((val1, val2))
            self.white_spinbox.spin_val.setValue(val2)

    def get_table(self, point_data, ind):
        xval = self.curve_plot.table_input
        if self.curve_plot.line_type == 'gamma':
            lims = (point_data[0, 0], point_data[-1, 0])
            table = gamma_line(xval, lims, self.gamma[ind], self.gray_max)
        else:
            table = self.curve_plot.update_table(point_data)
        return table

    def gamma_spinbox_changed(self):
        if self.line_type != 'gamma':
            return
        x_val = self.curve_plot.table_input
        for i in range(len(self.table_output)):
            if self.curve_plot.enable_channel[i]:
                self.gamma[i] = self.gamma_spinbox.spin_val.value()
                data = self.curve_plot.lut_points[i].data['pos'].copy()
                lims = (data[0, 0], data[-1, 0])
                table = gamma_line(x_val, lims, self.gamma[i], self.gray_max)
                self.table_output[i] = table.astype(int)
                self.curve_plot.set_lut_line(table, i)
                self.sig_table_changed.emit((self.table_output[i], i))

    def black_side_changed(self, diff_factor, ind):
        point_data = self.curve_plot.lut_points[ind].data['pos'].copy()
        da_width = point_data[-1, 0] - point_data[0, 0]
        dists = int(diff_factor * da_width)
        point_data[:-1, 0] = dists + point_data[:-1, 0]

        table = self.get_table(point_data, ind)
        self.table_output[ind] = table.astype(int)

        self.curve_plot.set_lut_points(point_data, ind)
        self.curve_plot.set_lut_line(table, ind)
        # self.white_spinbox.spin_val.setMinimum(point_data[-2, 0] + 1)

    def black_spinbox_changed(self):
        if self.curve_plot.hist_data is None:
            return
        val = self.black_spinbox.spin_val.value()
        val_diff = val - self.current_black_val
        diff_factor = val_diff / self.current_width
        for i in range(len(self.table_output)):
            if self.curve_plot.enable_channel[i]:
                self.black_side_changed(diff_factor, i)
                self.sig_table_changed.emit((self.table_output[i], i))
        self.current_black_val = val
        self.current_width = self.current_white_val - self.current_black_val

    def white_side_changed(self, diff_factor, ind):
        point_data = self.curve_plot.lut_points[ind].data['pos'].copy()
        da_width = point_data[-1, 0] - point_data[0, 0]
        dists = int(diff_factor * da_width)
        point_data[1:, 0] = dists + point_data[1:, 0]

        table = self.get_table(point_data, ind)
        self.table_output[ind] = table.astype(int)

        self.curve_plot.set_lut_points(point_data, ind)
        self.curve_plot.set_lut_line(table, ind)

    def white_spinbox_changed(self):
        if self.curve_plot.hist_data is None:
            return
        val = self.white_spinbox.spin_val.value()
        val_diff = val - self.current_white_val
        diff_factor = val_diff / self.current_width

        for i in range(len(self.table_output)):
            if self.curve_plot.enable_channel[i]:
                self.white_side_changed(diff_factor, i)
                self.sig_table_changed.emit((self.table_output[i], i))
        self.current_white_val = val
        self.current_width = self.current_white_val - self.current_black_val

    def table_changed(self, ev):
        table = ev[0]
        ind = ev[1]
        # curve_scatter = ev[1].astype(int)
        # self.black_spinbox.spin_val.setMaximum(curve_scatter[1, 0] - 1)
        # self.white_spinbox.spin_val.setMinimum(curve_scatter[-2, 0] + 1)
        self.sig_table_changed.emit((table, ind))

    def line_type_changed(self):
        self.line_type = self.line_type_combo.currentText()
        self.curve_plot.change_line_type(self.line_type)
        self.set_original_to_slider()
        if self.line_type != 'gamma':
            self.gamma_spinbox.setVisible(False)
        else:
            self.gamma_spinbox.setVisible(True)
            self.gamma_spinbox.spin_val.blockSignals(True)
            self.gamma_spinbox.spin_val.setValue(1)
            self.gamma_spinbox.spin_val.blockSignals(False)
        self.sig_line_type_changed.emit()

    def reset_pressed(self):
        self.curve_plot.change_line_type('gamma')
        self.set_original_to_slider()
        self.gamma_spinbox.spin_val.blockSignals(True)
        self.gamma_spinbox.spin_val.setValue(1)
        self.gamma_spinbox.setVisible(True)
        self.gamma_spinbox.spin_val.blockSignals(False)
        self.sig_reset.emit()

    def set_enable_induced_slider(self):
        all_start = []
        all_end = []
        for i in range(self.n_channels):
            all_start.append(self.curve_plot.lut_points[i].data['pos'][0, :])
            all_end.append(self.curve_plot.lut_points[i].data['pos'][-1, :])
        selected_start = np.asarray(all_start)[np.ravel(self.curve_plot.enable_channel)[:self.n_channels]]
        selected_end = np.asarray(all_end)[np.ravel(self.curve_plot.enable_channel)[:self.n_channels]]
        min_slider_val = int(np.min(selected_start[:, 0]))
        max_slider_val = int(np.max(selected_end[:, 0]))
        self.current_black_val = min_slider_val
        self.current_white_val = max_slider_val
        self.current_width = max_slider_val - min_slider_val
        self.block_signal(True)
        self.multi_handle_slider.setValue((min_slider_val, max_slider_val))
        self.black_spinbox.spin_val.setValue(min_slider_val)
        self.white_spinbox.spin_val.setValue(max_slider_val)
        self.block_signal(False)
        if np.sum(self.curve_plot.enable_channel) != 1:
            self.gamma_spinbox.spin_val.blockSignals(True)
            self.gamma_spinbox.spin_val.setValue(1)
            self.gamma_spinbox.spin_val.blockSignals(False)
        else:
            g_ind = np.where(self.curve_plot.enable_channel)[0][0]
            self.gamma_spinbox.spin_val.blockSignals(True)
            self.gamma_spinbox.spin_val.setValue(self.gamma[g_ind])
            self.gamma_spinbox.spin_val.blockSignals(False)

    def set_channel_enable(self, ind, is_enable):
        self.curve_plot.set_enable(ind, is_enable)
        self.set_enable_induced_slider()






